"""Tests for the legacy Protocol API module core implementations."""

import pytest
from decoy import Decoy

from opentrons.drivers.types import HeaterShakerLabwareLatchStatus
from opentrons.hardware_control import SyncHardwareAPI, SynchronousAdapter
from opentrons.hardware_control.modules import HeaterShaker
from opentrons.hardware_control.modules.types import (
    HeaterShakerModuleModel,
    SpeedStatus,
    TemperatureStatus,
)
from opentrons.hardware_control.types import Axis
from opentrons.protocol_api.core.legacy.legacy_module_core import (
    CannotPerformModuleAction,
    LegacyHeaterShakerCore,
    NoTargetTemperatureSetError,
    create_module_core,
)
from opentrons.protocol_api.core.legacy.legacy_protocol_core import (
    LegacyProtocolCore,
)
from opentrons.protocol_api.core.legacy.module_geometry import (
    HeaterShakerGeometry,
)
from opentrons.types import Location, Point

SyncHeaterShakerHardware = SynchronousAdapter[HeaterShaker]


@pytest.fixture
def mock_geometry(decoy: Decoy) -> HeaterShakerGeometry:
    """Get a mock heater-shaker geometry."""
    return decoy.mock(cls=HeaterShakerGeometry)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SyncHeaterShakerHardware:
    """Get a mock module hardware control interface."""
    return decoy.mock(name="SyncHeaterShakerHardware")  # type: ignore[no-any-return]


@pytest.fixture
def mock_sync_hardware_api(decoy: Decoy) -> SyncHardwareAPI:
    """Get a mock sync hardware API."""
    return decoy.mock(cls=SyncHardwareAPI)


@pytest.fixture
def mock_protocol_core(
    decoy: Decoy, mock_sync_hardware_api: SyncHardwareAPI
) -> LegacyProtocolCore:
    """Get a mock protocol core."""
    mock_protocol_core = decoy.mock(cls=LegacyProtocolCore)
    decoy.when(mock_protocol_core.get_hardware()).then_return(mock_sync_hardware_api)
    return mock_protocol_core


@pytest.fixture
def subject(
    mock_geometry: HeaterShakerGeometry,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    mock_protocol_core: LegacyProtocolCore,
) -> LegacyHeaterShakerCore:
    """Get a legacy module implementation core with mocked out dependencies."""
    return LegacyHeaterShakerCore(
        requested_model=HeaterShakerModuleModel.HEATER_SHAKER_V1,
        geometry=mock_geometry,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_geometry: HeaterShakerGeometry,
    mock_protocol_core: LegacyProtocolCore,
) -> None:
    """It should be able to create a magnetic module core."""
    mock_module_hardware_api = decoy.mock(cls=HeaterShaker)
    result = create_module_core(
        geometry=mock_geometry,
        module_hardware_api=mock_module_hardware_api,
        requested_model=HeaterShakerModuleModel.HEATER_SHAKER_V1,
        protocol_core=mock_protocol_core,
    )

    assert isinstance(result, LegacyHeaterShakerCore)


def test_get_current_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should report the current temperature."""
    decoy.when(mock_sync_module_hardware.temperature).then_return(42.0)
    result = subject.get_current_temperature()
    assert result == 42.0


def test_get_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should report the target temperature."""
    decoy.when(mock_sync_module_hardware.target_temperature).then_return(42.0)
    result = subject.get_target_temperature()
    assert result == 42.0


def test_get_temperature_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should report the temperature status."""
    decoy.when(mock_sync_module_hardware.temperature_status).then_return(
        TemperatureStatus.COOLING
    )
    result = subject.get_temperature_status()
    assert result == TemperatureStatus.COOLING


def test_get_current_speed(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should report the current speed."""
    decoy.when(mock_sync_module_hardware.speed).then_return(321)
    result = subject.get_current_speed()
    assert result == 321


def test_get_target_speed(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should report the target speed."""
    decoy.when(mock_sync_module_hardware.target_speed).then_return(321)
    result = subject.get_target_speed()
    assert result == 321


def test_get_speed_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should report the speed status."""
    decoy.when(mock_sync_module_hardware.speed_status).then_return(
        SpeedStatus.ACCELERATING
    )
    result = subject.get_speed_status()
    assert result == SpeedStatus.ACCELERATING


def test_get_labware_latch_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should report the labware latch status."""
    decoy.when(mock_sync_module_hardware.labware_latch_status).then_return(
        HeaterShakerLabwareLatchStatus.OPENING
    )
    result = subject.get_labware_latch_status()
    assert result == HeaterShakerLabwareLatchStatus.OPENING


def test_set_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should set the target temperature with the hardware."""
    subject.set_target_temperature(42.0)

    decoy.verify(mock_sync_module_hardware.start_set_temperature(42.0), times=1)


def test_wait_for_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should wait for the target temperature with the hardware."""
    decoy.when(subject.get_target_temperature()).then_return(42.0)

    subject.wait_for_target_temperature()

    decoy.verify(mock_sync_module_hardware.await_temperature(42.0), times=1)


def test_wait_for_temperature_no_target(
    decoy: Decoy,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should raise a NoTargetTemperatureSetError."""
    decoy.when(subject.get_target_temperature()).then_return(None)

    with pytest.raises(NoTargetTemperatureSetError):
        subject.wait_for_target_temperature()


def test_set_and_wait_for_shake_speed(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    mock_geometry: HeaterShakerGeometry,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_protocol_core: LegacyProtocolCore,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should set and wait for the target speed with the hardware."""
    decoy.when(subject.get_labware_latch_status()).then_return(
        HeaterShakerLabwareLatchStatus.IDLE_CLOSED
    )
    decoy.when(mock_protocol_core.get_last_location()).then_return(
        Location(point=Point(x=1, y=2, z=3), labware=None)
    )
    decoy.when(
        mock_geometry.is_pipette_blocking_shake_movement(
            Location(point=Point(x=1, y=2, z=3), labware=None)
        )
    ).then_return(True)

    subject.set_and_wait_for_shake_speed(1337)

    decoy.verify(
        mock_sync_hardware_api.home(axes=[Axis.Z, Axis.A]),
        mock_protocol_core.set_last_location(None),
        mock_sync_module_hardware.set_speed(rpm=1337),
    )


def test_set_and_wait_for_shake_speed_no_prep(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    mock_geometry: HeaterShakerGeometry,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_protocol_core: LegacyProtocolCore,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should set and wait for the target speed with the hardware."""
    decoy.when(subject.get_labware_latch_status()).then_return(
        HeaterShakerLabwareLatchStatus.IDLE_CLOSED
    )
    decoy.when(mock_protocol_core.get_last_location()).then_return(
        Location(point=Point(x=1, y=2, z=3), labware=None)
    )
    decoy.when(
        mock_geometry.is_pipette_blocking_shake_movement(
            Location(point=Point(x=1, y=2, z=3), labware=None)
        )
    ).then_return(False)

    subject.set_and_wait_for_shake_speed(1337)

    decoy.verify(mock_sync_hardware_api.home(axes=[Axis.Z, Axis.A]), times=0)
    decoy.verify(mock_sync_module_hardware.set_speed(rpm=1337), times=1)


@pytest.mark.parametrize(
    "latch_status",
    [
        HeaterShakerLabwareLatchStatus.IDLE_UNKNOWN,
        HeaterShakerLabwareLatchStatus.UNKNOWN,
        HeaterShakerLabwareLatchStatus.CLOSING,
        HeaterShakerLabwareLatchStatus.IDLE_OPEN,
        HeaterShakerLabwareLatchStatus.OPENING,
    ],
)
def test_set_and_wait_for_shake_speed_raises(
    decoy: Decoy,
    latch_status: HeaterShakerLabwareLatchStatus,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should raise a CannotPerformModuleAction when latch is not closed."""
    decoy.when(subject.get_labware_latch_status()).then_return(latch_status)

    with pytest.raises(CannotPerformModuleAction):
        subject.set_and_wait_for_shake_speed(rpm=1337)


def test_open_labware_latch(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    mock_geometry: HeaterShakerGeometry,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_protocol_core: LegacyProtocolCore,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should open the labware latch with the hardware."""
    decoy.when(subject.get_speed_status()).then_return(SpeedStatus.IDLE)
    decoy.when(mock_protocol_core.get_last_location()).then_return(
        Location(point=Point(x=1, y=2, z=3), labware=None)
    )
    decoy.when(
        mock_geometry.is_pipette_blocking_latch_movement(
            Location(point=Point(x=1, y=2, z=3), labware=None)
        )
    ).then_return(True)

    subject.open_labware_latch()

    decoy.verify(
        mock_sync_hardware_api.home(axes=[Axis.Z, Axis.A]),
        mock_protocol_core.set_last_location(None),
        mock_sync_module_hardware.open_labware_latch(),
    )


def test_open_labware_latch_no_prep(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    mock_geometry: HeaterShakerGeometry,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_protocol_core: LegacyProtocolCore,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should open the labware latch with the hardware."""
    decoy.when(subject.get_speed_status()).then_return(SpeedStatus.IDLE)
    decoy.when(mock_protocol_core.get_last_location()).then_return(
        Location(point=Point(x=1, y=2, z=3), labware=None)
    )
    decoy.when(
        mock_geometry.is_pipette_blocking_latch_movement(
            Location(point=Point(x=1, y=2, z=3), labware=None)
        )
    ).then_return(False)

    subject.open_labware_latch()

    decoy.verify(mock_sync_hardware_api.home(axes=[Axis.Z, Axis.A]), times=0)
    decoy.verify(
        mock_sync_module_hardware.open_labware_latch(),
        times=1,
    )


@pytest.mark.parametrize(
    "speed_status",
    [
        SpeedStatus.DECELERATING,
        SpeedStatus.ACCELERATING,
        SpeedStatus.HOLDING,
        SpeedStatus.ERROR,
    ],
)
def test_open_labware_latch_raises(
    decoy: Decoy,
    speed_status: SpeedStatus,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should raise a CannotPerformModuleAction when heater-shaker is not idle."""
    decoy.when(subject.get_speed_status()).then_return(speed_status)

    with pytest.raises(CannotPerformModuleAction):
        subject.open_labware_latch()


def test_close_labware_latch(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should close the labware latch with the hardware."""
    subject.close_labware_latch()

    decoy.verify(mock_sync_module_hardware.close_labware_latch(), times=1)


def test_deactivate_shaker(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should stop shaking with the hardware."""
    subject.deactivate_shaker()

    decoy.verify(mock_sync_module_hardware.deactivate_shaker(), times=1)


def test_deactivate_heater(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: LegacyHeaterShakerCore,
) -> None:
    """It should stop heating with the hardware."""
    subject.deactivate_heater()

    decoy.verify(mock_sync_module_hardware.deactivate_heater(), times=1)



from opentrons import protocol_api

requirements = {"robotType": "OT-2", "apiLevel": "2.16"}


def run(ctx: protocol_api.ProtocolContext) -> None:
    tipracks = [ctx.load_labware("opentrons_96_tiprack_300ul", "5")]
    m300 = ctx.load_instrument("p300_multi_gen2", "right", tipracks)

    m300.pick_up_tip()
    m300.drop_tip()



"""Test for the ProtocolEngine-based protocol API core."""

import inspect
from typing import Optional, Tuple, Type, cast

import pytest
from decoy import Decoy

from opentrons_shared_data import liquid_classes
from opentrons_shared_data.deck import load as load_deck
from opentrons_shared_data.deck.types import (
    DeckDefinitionV5,
    SlotDefV3,
)
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition2,
    labware_definition_type_adapter,
)
from opentrons_shared_data.labware.types import (
    LabwareDefinition as LabwareDefDict,
)
from opentrons_shared_data.labware.types import (
    LabwareUri,
)
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
)
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.robot.types import RobotType

from ... import versions_at_or_above, versions_below
from opentrons.hardware_control import SyncHardwareAPI, SynchronousAdapter
from opentrons.hardware_control.modules import AbstractModule
from opentrons.hardware_control.modules.types import (
    FlexStackerModuleModel,
    HeaterShakerModuleModel,
    MagneticBlockModel,
    MagneticModuleModel,
    ModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
)
from opentrons.protocol_api import MAX_SUPPORTED_VERSION, OFF_DECK, validation
from opentrons.protocol_api._liquid import Liquid, LiquidClass
from opentrons.protocol_api.core.engine import (
    InstrumentCore,
    LabwareCore,
    ModuleCore,
    ProtocolCore,
    _default_liquid_class_versions,
    deck_conflict,
    load_labware_params,
)
from opentrons.protocol_api.core.engine.exceptions import InvalidModuleLocationError
from opentrons.protocol_api.core.engine.module_core import (
    HeaterShakerModuleCore,
    MagneticModuleCore,
    NonConnectedModuleCore,
    TemperatureModuleCore,
    ThermocyclerModuleCore,
)
from opentrons.protocol_api.core.engine.tasks import EngineTaskCore
from opentrons.protocol_api.core.labware import LabwareLoadParams
from opentrons.protocol_api.disposal_locations import TrashBin, WasteChute
from opentrons.protocol_engine import (
    AddressableAreaLocation,
    DeckSlotLocation,
    LabwareMovementStrategy,
    LabwareOffsetVector,
    LoadedLabware,
    LoadedModule,
    ModuleLocation,
    OnLabwareLocation,
    commands,
)
from opentrons.protocol_engine import (
    ModuleModel as EngineModuleModel,
)
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocol_engine.errors import (
    LabwareNotLoadedOnLabwareError,
    LabwareNotLoadedOnModuleError,
)
from opentrons.protocol_engine.state.labware import (
    LabwareLoadParams as EngineLabwareLoadParams,
)
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    FlowRates,
    HexColor,
)
from opentrons.protocol_engine.types import (
    Liquid as PE_Liquid,
)
from opentrons.protocols.api_support.deck_type import (
    STANDARD_OT2_DECK,
    STANDARD_OT3_DECK,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import DeckSlotName, Mount, MountType, Point, StagingSlotName


@pytest.fixture(scope="session")
def ot2_standard_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(STANDARD_OT2_DECK, 5)


@pytest.fixture(scope="session")
def ot3_standard_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(STANDARD_OT3_DECK, 5)


@pytest.fixture(autouse=True)
def patch_mock_load_labware_params(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out load_labware_params.py functions."""
    for name, func in inspect.getmembers(load_labware_params, inspect.isfunction):
        monkeypatch.setattr(load_labware_params, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def patch_default_liquid_class_versions(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out _default_liquid_class_versions.py functions."""
    for name, func in inspect.getmembers(
        _default_liquid_class_versions, inspect.isfunction
    ):
        monkeypatch.setattr(_default_liquid_class_versions, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def patch_mock_load_liquid_class_def(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out liquid_classes.load_definition() function."""
    mock_load = decoy.mock(func=liquid_classes.load_definition)
    monkeypatch.setattr(liquid_classes, "load_definition", mock_load)


@pytest.fixture(autouse=True)
def patch_mock_validation(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out validation.py functions."""
    for name, func in inspect.getmembers(validation, inspect.isfunction):
        monkeypatch.setattr(validation, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def patch_mock_deck_conflict_check(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Replace deck_conflict.check() with a mock."""
    mock = decoy.mock(func=deck_conflict.check)
    monkeypatch.setattr(deck_conflict, "check", mock)


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def api_version() -> APIVersion:
    """Get an API version to apply to the interface."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SynchronousAdapter[AbstractModule]:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="SynchronousAdapter[AbstractModule]")  # type: ignore[no-any-return]


@pytest.fixture
def mock_sync_hardware_api(decoy: Decoy) -> SyncHardwareAPI:
    """Get a mock hardware API."""
    return decoy.mock(cls=SyncHardwareAPI)


@pytest.fixture
def subject(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    api_version: APIVersion,
    mock_sync_hardware_api: SyncHardwareAPI,
    request: pytest.FixtureRequest,
) -> ProtocolCore:
    """Get a ProtocolCore test subject with its dependencies mocked out."""
    if (
        hasattr(request.node, "callspec")
        and "api_version" in request.node.callspec.params
    ):
        selected_api_version = request.node.callspec.params["api_version"]
    else:
        selected_api_version = api_version
    decoy.when(mock_engine_client.state.labware.get_fixed_trash_id()).then_return(
        "fixed-trash-123"
    )
    decoy.when(
        mock_engine_client.state.labware.get_definition("fixed-trash-123")
    ).then_return(
        LabwareDefinition2.model_construct(ordering=[["A1"]])  # type: ignore[call-arg]
    )

    return ProtocolCore(
        engine_client=mock_engine_client,
        api_version=selected_api_version,
        sync_hardware=mock_sync_hardware_api,
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 3)])
def test_api_version(
    decoy: Decoy, subject: ProtocolCore, api_version: APIVersion
) -> None:
    """Should return the protocol version."""
    assert subject.api_version == api_version


def test_get_slot_definition(
    ot2_standard_deck_def: DeckDefinitionV5, subject: ProtocolCore, decoy: Decoy
) -> None:
    """It should return a deck slot's definition."""
    expected_slot_def = cast(
        SlotDefV3,
        {
            "id": "abc",
            "position": [1, 2, 3],
            "boundingBox": {
                "xDimension": 4,
                "yDimension": 5,
                "zDimension": 6,
            },
            "displayName": "xyz",
            "compatibleModuleTypes": [],
        },
    )
    decoy.when(
        subject._engine_client.state.addressable_areas.get_slot_definition(
            DeckSlotName.SLOT_6.id
        )
    ).then_return(expected_slot_def)
    assert subject.get_slot_definition(DeckSlotName.SLOT_6) == expected_slot_def


# APIv2.15 because we're expecting a fixed trash.
@pytest.mark.parametrize("api_version", [APIVersion(2, 15)])
def test_fixed_trash(subject: ProtocolCore) -> None:
    """It should have a single labware core for the fixed trash."""
    result = subject.fixed_trash

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "fixed-trash-123"
    assert subject.get_labware_cores() == [result]

    # verify it's the same core every time
    assert subject.fixed_trash is result


def test_get_slot_item_empty(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should return None for an empty deck slot."""
    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=DeckSlotName.SLOT_1,
        )
    ).then_return(None)

    assert subject.get_slot_item(DeckSlotName.SLOT_1) is None


@pytest.mark.parametrize(
    "api_version", versions_below(APIVersion(2, 19), flex_only=False)
)
def test_load_instrument_pre_219(
    decoy: Decoy,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadPipette command."""
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadPipetteParams(
                pipetteName=PipetteNameType.P300_SINGLE,
                mount=MountType.LEFT,
                tipOverlapNotAfterVersion="v0",
                liquidPresenceDetection=False,
            )
        )
    ).then_return(commands.LoadPipetteResult(pipetteId="cool-pipette"))

    decoy.when(
        mock_engine_client.state.pipettes.get_flow_rates("cool-pipette")
    ).then_return(
        FlowRates(
            default_aspirate={"1.1": 22},
            default_dispense={"3.3": 44},
            default_blow_out={"5.5": 66},
        ),
    )

    result = subject.load_instrument(
        instrument_name=PipetteNameType.P300_SINGLE, mount=Mount.LEFT
    )

    assert isinstance(result, InstrumentCore)
    assert result.pipette_id == "cool-pipette"


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 20)))
def test_load_instrument_post_220(
    decoy: Decoy,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadPipette command."""
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadPipetteParams(
                pipetteName=PipetteNameType.P300_SINGLE,
                mount=MountType.LEFT,
                tipOverlapNotAfterVersion="v3",
                liquidPresenceDetection=False,
            )
        )
    ).then_return(commands.LoadPipetteResult(pipetteId="cool-pipette"))

    decoy.when(
        mock_engine_client.state.pipettes.get_flow_rates("cool-pipette")
    ).then_return(
        FlowRates(
            default_aspirate={"1.1": 22},
            default_dispense={"3.3": 44},
            default_blow_out={"5.5": 66},
        ),
    )

    result = subject.load_instrument(
        instrument_name=PipetteNameType.P300_SINGLE, mount=Mount.LEFT
    )

    assert isinstance(result, InstrumentCore)
    assert result.pipette_id == "cool-pipette"


def test_load_labware(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadLabware command."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_labware",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
                loadName="some_labware",
                displayName="some_display_name",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLabwareResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            offsetId=None,
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    result = subject.load_labware(
        load_name="some_labware",
        location=DeckSlotName.SLOT_5,
        label="some_display_name",  # maps to optional display name
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"
    assert subject.get_labware_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=DeckSlotName.SLOT_5,
        )
    ).then_return(
        LoadedLabware.model_construct(id="abc123")  # type: ignore[call-arg]
    )

    assert subject.get_slot_item(DeckSlotName.SLOT_5) is result


def test_load_labware_on_staging_slot(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadLabware command for a labware on a staging slot."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_labware",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                location=AddressableAreaLocation(addressableAreaName="B4"),
                loadName="some_labware",
                displayName="some_display_name",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLabwareResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            offsetId=None,
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    result = subject.load_labware(
        load_name="some_labware",
        location=StagingSlotName.SLOT_B4,
        label="some_display_name",  # maps to optional display name
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"
    assert subject.get_labware_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=StagingSlotName.SLOT_B4,
        )
    ).then_return(
        LoadedLabware.model_construct(id="abc123")  # type: ignore[call-arg]
    )

    assert subject.get_slot_item(StagingSlotName.SLOT_B4) is result


def test_load_labware_on_labware(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadLabware command onto an OnLabware location."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_labware_core.labware_id).then_return("labware-id")

    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_labware",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                location=OnLabwareLocation(labwareId="labware-id"),
                loadName="some_labware",
                displayName="some_display_name",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLabwareResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            offsetId=None,
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    decoy.when(
        mock_engine_client.state.labware.get_id_by_labware("labware-id")
    ).then_return("abc123")

    result = subject.load_labware(
        load_name="some_labware",
        location=mock_labware_core,
        label="some_display_name",
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"
    assert subject.get_labware_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    assert subject.get_labware_on_labware(mock_labware_core) is result


def test_load_labware_off_deck(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadLabware off deck command."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_labware",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                location=OFF_DECK_LOCATION,
                loadName="some_labware",
                displayName="some_display_name",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLabwareResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            offsetId=None,
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    result = subject.load_labware(
        load_name="some_labware",
        location=OFF_DECK,
        label="some_display_name",  # maps to optional display name
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"
    assert subject.get_labware_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )


def test_load_adapter(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadLabware command for an adapter."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_adapter",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
                loadName="some_adapter",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLabwareResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            offsetId=None,
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    result = subject.load_adapter(
        load_name="some_adapter",
        location=DeckSlotName.SLOT_5,
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"
    assert subject.get_labware_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=DeckSlotName.SLOT_5,
        )
    ).then_return(
        LoadedLabware.model_construct(id="abc123")  # type: ignore[call-arg]
    )

    assert subject.get_slot_item(DeckSlotName.SLOT_5) is result


def test_load_adapter_on_staging_slot(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadLabware command for an adapter."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_adapter",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                location=AddressableAreaLocation(addressableAreaName="B4"),
                loadName="some_adapter",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLabwareResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            offsetId=None,
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    result = subject.load_adapter(
        load_name="some_adapter",
        location=StagingSlotName.SLOT_B4,
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"
    assert subject.get_labware_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=StagingSlotName.SLOT_B4,
        )
    ).then_return(
        LoadedLabware.model_construct(id="abc123")  # type: ignore[call-arg]
    )

    assert subject.get_slot_item(StagingSlotName.SLOT_B4) is result


def test_load_lid(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadLid command."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_labware_core.labware_id).then_return("labware-id")
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_labware",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLidParams(
                location=OnLabwareLocation(labwareId="labware-id"),
                loadName="some_labware",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLidResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(ordering=[]),  # type: ignore[call-arg]
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    result = subject.load_lid(
        load_name="some_labware",
        location=mock_labware_core,
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"
    assert subject.get_labware_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=DeckSlotName.SLOT_5,
        )
    ).then_return(
        LoadedLabware.model_construct(id="abc123")  # type: ignore[call-arg]
    )

    assert subject.get_slot_item(DeckSlotName.SLOT_5) is result


def test_load_lid_stack(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a LoadLidStack command."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_labware",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLidStackParams(
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
                loadName="some_labware",
                namespace="some_namespace",
                version=9001,
                quantity=5,
            )
        )
    ).then_return(
        commands.LoadLidStackResult.model_construct(
            stackLabwareId="abc123",
            labwareIds=["1", "2", "3", "4", "5"],
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    result = subject.load_lid_stack(
        load_name="some_labware",
        location=DeckSlotName.SLOT_5,
        namespace="a_namespace",
        version=456,
        quantity=5,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"
    assert subject.get_labware_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=DeckSlotName.SLOT_5,
        )
    ).then_return(
        LoadedLabware.model_construct(id="abc123")  # type: ignore[call-arg]
    )

    assert subject.get_slot_item(DeckSlotName.SLOT_5) is result


def test_load_trash_bin(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should load a trash bin."""
    prior_disposal_locations = subject.get_disposal_locations()
    trash = subject.load_trash_bin(
        slot_name=DeckSlotName.SLOT_D2, area_name="my trendy area"
    )
    assert isinstance(trash, TrashBin)
    decoy.verify(
        mock_engine_client.state.addressable_areas.raise_if_area_not_in_deck_configuration(
            "my trendy area"
        ),
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            new_trash_bin=trash,
            existing_disposal_locations=prior_disposal_locations,
            existing_labware_ids=[],
            existing_module_ids=[],
        ),
        mock_engine_client.add_addressable_area("my trendy area"),
    )

    assert trash in subject.get_disposal_locations()


def test_load_ot2_fixed_trash_bin(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should load a fixed trash bin for the OT-2."""
    prior_disposal_locations = subject.get_disposal_locations()
    subject.load_ot2_fixed_trash_bin()
    fixed_trash = subject.get_disposal_locations()[-1]
    assert isinstance(fixed_trash, TrashBin)
    assert fixed_trash.area_name == "fixedTrash"
    decoy.verify(
        mock_engine_client.state.addressable_areas.raise_if_area_not_in_deck_configuration(
            "fixedTrash"
        ),
        times=0,
    )
    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            new_trash_bin=fixed_trash,
            existing_disposal_locations=prior_disposal_locations,
            existing_labware_ids=[],
            existing_module_ids=[],
        ),
        times=0,
    )
    decoy.verify(mock_engine_client.add_addressable_area("fixedTrash"), times=0)


def test_load_waste_chute(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should load a waste chute."""
    waste_chute = subject.load_waste_chute()
    assert isinstance(waste_chute, WasteChute)
    decoy.verify(
        mock_engine_client.state.addressable_areas.raise_if_area_not_in_deck_configuration(
            "1ChannelWasteChute"
        ),
        mock_engine_client.add_addressable_area("1ChannelWasteChute"),
    )

    assert waste_chute in subject.get_disposal_locations()


@pytest.mark.parametrize(
    argnames=["use_gripper", "pause_for_manual_move", "expected_strategy"],
    argvalues=[
        (True, False, LabwareMovementStrategy.USING_GRIPPER),
        (True, True, LabwareMovementStrategy.USING_GRIPPER),
        (False, False, LabwareMovementStrategy.MANUAL_MOVE_WITHOUT_PAUSE),
        (False, True, LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE),
    ],
)
@pytest.mark.parametrize(
    argnames=["pick_up_offset", "drop_offset"],
    argvalues=[
        (None, None),
        (None, (4, 5, 6)),
        ((4, 5, 6), (4, 5, 6)),
    ],
)
def test_move_labware(
    decoy: Decoy,
    subject: ProtocolCore,
    mock_engine_client: EngineClient,
    expected_strategy: LabwareMovementStrategy,
    use_gripper: bool,
    pause_for_manual_move: bool,
    pick_up_offset: Optional[Tuple[float, float, float]],
    drop_offset: Optional[Tuple[float, float, float]],
) -> None:
    """It should issue a move labware command to the engine."""
    decoy.when(
        mock_engine_client.state.labware.get_definition("labware-id")
    ).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )
    labware = LabwareCore(labware_id="labware-id", engine_client=mock_engine_client)
    subject.move_labware(
        labware_core=labware,
        new_location=DeckSlotName.SLOT_5,
        use_gripper=use_gripper,
        pause_for_manual_move=pause_for_manual_move,
        pick_up_offset=pick_up_offset,
        drop_offset=drop_offset,
    )
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.MoveLabwareParams(
                labwareId="labware-id",
                newLocation=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
                strategy=expected_strategy,
                pickUpOffset=(
                    LabwareOffsetVector(x=4, y=5, z=6) if pick_up_offset else None
                ),
                dropOffset=LabwareOffsetVector(x=4, y=5, z=6) if drop_offset else None,
            )
        ),
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="labware-id",
        ),
    )


def test_move_labware_on_staging_slot(
    decoy: Decoy,
    subject: ProtocolCore,
    mock_engine_client: EngineClient,
    api_version: APIVersion,
) -> None:
    """It should issue a move labware command to the engine."""
    decoy.when(
        mock_engine_client.state.labware.get_definition("labware-id")
    ).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )
    labware = LabwareCore(labware_id="labware-id", engine_client=mock_engine_client)
    subject.move_labware(
        labware_core=labware,
        new_location=StagingSlotName.SLOT_B4,
        use_gripper=False,
        pause_for_manual_move=True,
        pick_up_offset=None,
        drop_offset=None,
    )
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.MoveLabwareParams(
                labwareId="labware-id",
                newLocation=AddressableAreaLocation(addressableAreaName="B4"),
                strategy=LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE,
                pickUpOffset=None,
                dropOffset=None,
            )
        ),
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="labware-id",
        ),
    )


def test_move_labware_on_non_connected_module(
    decoy: Decoy,
    subject: ProtocolCore,
    mock_engine_client: EngineClient,
    api_version: APIVersion,
) -> None:
    """It should issue a move labware command to the engine."""
    decoy.when(
        mock_engine_client.state.labware.get_definition("labware-id")
    ).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )
    labware = LabwareCore(labware_id="labware-id", engine_client=mock_engine_client)
    non_connected_module_core = NonConnectedModuleCore(
        module_id="module-id",
        engine_client=mock_engine_client,
        api_version=api_version,
        protocol_core=subject,
    )
    subject.move_labware(
        labware_core=labware,
        new_location=non_connected_module_core,
        use_gripper=False,
        pause_for_manual_move=True,
        pick_up_offset=None,
        drop_offset=None,
    )
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.MoveLabwareParams(
                labwareId="labware-id",
                newLocation=ModuleLocation(moduleId="module-id"),
                strategy=LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE,
                pickUpOffset=None,
                dropOffset=None,
            )
        ),
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="labware-id",
        ),
    )


def test_move_labware_off_deck(
    decoy: Decoy,
    subject: ProtocolCore,
    mock_engine_client: EngineClient,
    api_version: APIVersion,
) -> None:
    """It should issue a move labware command to the engine with pause strategy."""
    decoy.when(
        mock_engine_client.state.labware.get_definition("labware-id")
    ).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )
    labware = LabwareCore(labware_id="labware-id", engine_client=mock_engine_client)

    subject.move_labware(
        labware_core=labware,
        new_location=OFF_DECK,
        use_gripper=False,
        pause_for_manual_move=True,
        pick_up_offset=None,
        drop_offset=None,
    )
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.MoveLabwareParams(
                labwareId="labware-id",
                newLocation=OFF_DECK_LOCATION,
                strategy=LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE,
                pickUpOffset=None,
                dropOffset=None,
            )
        ),
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="labware-id",
        ),
    )


def test_load_labware_on_module(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SynchronousAdapter[AbstractModule],
    subject: ProtocolCore,
    api_version: APIVersion,
) -> None:
    """It should issue a LoadLabware command."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_labware",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                location=ModuleLocation(moduleId="module-id"),
                loadName="some_labware",
                displayName="some_display_name",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLabwareResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            offsetId=None,
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    module_core = ModuleCore(
        module_id="module-id",
        engine_client=mock_engine_client,
        api_version=api_version,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=subject,
    )

    result = subject.load_labware(
        load_name="some_labware",
        location=module_core,
        label="some_display_name",  # maps to optional display name
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.labware.get_id_by_module("module-id")
    ).then_return("abc123")

    assert subject.get_labware_on_module(module_core) is result


def test_load_labware_on_non_connected_module(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
    api_version: APIVersion,
) -> None:
    """It should issue a LoadLabware command."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return([EngineLabwareLoadParams("hello", "world", 654)])

    decoy.when(
        load_labware_params.resolve(
            "some_labware",
            "a_namespace",
            456,
            [EngineLabwareLoadParams("hello", "world", 654)],
            subject.api_version,
        )
    ).then_return(("some_namespace", 9001))

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                location=ModuleLocation(moduleId="module-id"),
                loadName="some_labware",
                displayName="some_display_name",
                namespace="some_namespace",
                version=9001,
            )
        )
    ).then_return(
        commands.LoadLabwareResult.model_construct(
            labwareId="abc123",
            definition=LabwareDefinition2.model_construct(),  # type: ignore[call-arg]
            offsetId=None,
        )
    )

    decoy.when(mock_engine_client.state.labware.get_definition("abc123")).then_return(
        LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]
    )

    non_connected_module_core = NonConnectedModuleCore(
        module_id="module-id",
        engine_client=mock_engine_client,
        api_version=api_version,
        protocol_core=subject,
    )

    result = subject.load_labware(
        load_name="some_labware",
        location=non_connected_module_core,
        label="some_display_name",  # maps to optional display name
        namespace="a_namespace",
        version=456,
    )

    assert isinstance(result, LabwareCore)
    assert result.labware_id == "abc123"

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_labware_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.labware.get_id_by_module("module-id")
    ).then_return("abc123")

    assert subject.get_labware_on_module(non_connected_module_core) is result


def test_add_labware_definition(
    decoy: Decoy,
    minimal_labware_def: LabwareDefDict,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should add a labware definition to the engine."""
    decoy.when(
        mock_engine_client.add_labware_definition(
            definition=labware_definition_type_adapter.validate_python(
                minimal_labware_def
            )
        )
    ).then_return(LabwareUri("hello/world/123"))

    result = subject.add_labware_definition(minimal_labware_def)

    assert result == LabwareLoadParams("hello", "world", 123)


@pytest.mark.parametrize(
    (
        "requested_model",
        "engine_model",
        "expected_core_cls",
        "slot_name",
        "robot_type",
    ),
    [
        (
            TemperatureModuleModel.TEMPERATURE_V1,
            EngineModuleModel.TEMPERATURE_MODULE_V1,
            TemperatureModuleCore,
            DeckSlotName.SLOT_1,
            "OT-2 Standard",
        ),
        (
            TemperatureModuleModel.TEMPERATURE_V2,
            EngineModuleModel.TEMPERATURE_MODULE_V2,
            TemperatureModuleCore,
            DeckSlotName.SLOT_D1,
            "OT-3 Standard",
        ),
        (
            MagneticModuleModel.MAGNETIC_V1,
            EngineModuleModel.MAGNETIC_MODULE_V1,
            MagneticModuleCore,
            DeckSlotName.SLOT_1,
            "OT-2 Standard",
        ),
        (
            ThermocyclerModuleModel.THERMOCYCLER_V1,
            EngineModuleModel.THERMOCYCLER_MODULE_V1,
            ThermocyclerModuleCore,
            DeckSlotName.SLOT_7,
            "OT-2 Standard",
        ),
        (
            ThermocyclerModuleModel.THERMOCYCLER_V2,
            EngineModuleModel.THERMOCYCLER_MODULE_V2,
            ThermocyclerModuleCore,
            DeckSlotName.SLOT_B1,
            "OT-3 Standard",
        ),
        (
            HeaterShakerModuleModel.HEATER_SHAKER_V1,
            EngineModuleModel.HEATER_SHAKER_MODULE_V1,
            HeaterShakerModuleCore,
            DeckSlotName.SLOT_A1,
            "OT-3 Standard",
        ),
    ],
)
# APIv2.15 because we're expecting a fixed trash.
@pytest.mark.parametrize("api_version", [APIVersion(2, 15)])
def test_load_module(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_hardware_api: SyncHardwareAPI,
    requested_model: ModuleModel,
    engine_model: EngineModuleModel,
    expected_core_cls: Type[ModuleCore],
    subject: ProtocolCore,
    slot_name: DeckSlotName,
    robot_type: RobotType,
) -> None:
    """It should issue a load module engine command."""
    mock_hw_mod_1 = decoy.mock(cls=AbstractModule)
    mock_hw_mod_2 = decoy.mock(cls=AbstractModule)

    decoy.when(mock_hw_mod_1.device_info).then_return({"serial": "abc123"})
    decoy.when(mock_hw_mod_2.device_info).then_return({"serial": "xyz789"})
    decoy.when(mock_sync_hardware_api.attached_modules).then_return(
        [mock_hw_mod_1, mock_hw_mod_2]
    )

    decoy.when(mock_engine_client.state.config.robot_type).then_return(robot_type)

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadModuleParams(
                model=engine_model,
                location=DeckSlotLocation(slotName=slot_name),
            )
        )
    ).then_return(
        commands.LoadModuleResult(
            moduleId="abc123",
            model=engine_model,
            serialNumber="xyz789",
        )
    )

    result = subject.load_module(
        model=requested_model,
        deck_slot=slot_name,
        configuration=None,
    )

    assert isinstance(result, expected_core_cls)
    assert result.module_id == "abc123"
    assert subject.get_module_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=["fixed-trash-123"],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_module_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=slot_name,
        )
    ).then_return(
        LoadedModule.model_construct(id="abc123")  # type: ignore[call-arg]
    )
    decoy.when(mock_engine_client.state.labware.get_id_by_module("abc123")).then_raise(
        LabwareNotLoadedOnModuleError("oh no")
    )

    assert subject.get_slot_item(slot_name) is result
    assert subject.get_labware_on_module(result) is None


# APIv2.15 because we're expecting a fixed trash.
@pytest.mark.parametrize("api_version", [APIVersion(2, 15)])
def test_load_mag_block(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a load module engine command."""
    decoy.when(mock_engine_client.state.config.robot_type).then_return("OT-3 Standard")

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadModuleParams(
                model=EngineModuleModel.MAGNETIC_BLOCK_V1,
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A2),
            )
        )
    ).then_return(
        commands.LoadModuleResult(
            moduleId="abc123",
            model=EngineModuleModel.MAGNETIC_BLOCK_V1,
            serialNumber=None,
        )
    )

    result = subject.load_module(
        model=MagneticBlockModel.MAGNETIC_BLOCK_V1,
        deck_slot=DeckSlotName.SLOT_A2,
        configuration=None,
    )

    assert isinstance(result, NonConnectedModuleCore)
    assert result.module_id == "abc123"
    assert subject.get_module_cores() == [result]

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=["fixed-trash-123"],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_module_id="abc123",
        )
    )

    decoy.when(
        mock_engine_client.state.geometry.get_slot_item(
            slot_name=DeckSlotName.SLOT_1,
        )
    ).then_return(
        LoadedModule.model_construct(id="abc123")  # type: ignore[call-arg]
    )
    decoy.when(mock_engine_client.state.labware.get_id_by_module("abc123")).then_raise(
        LabwareNotLoadedOnModuleError("oh no")
    )

    assert subject.get_slot_item(DeckSlotName.SLOT_1) is result
    assert subject.get_labware_on_module(result) is None


@pytest.mark.parametrize(
    ("requested_model", "engine_model", "expected_slot"),
    [
        (
            ThermocyclerModuleModel.THERMOCYCLER_V1,
            EngineModuleModel.THERMOCYCLER_MODULE_V1,
            DeckSlotName.SLOT_B1,
        ),
        (
            ThermocyclerModuleModel.THERMOCYCLER_V2,
            EngineModuleModel.THERMOCYCLER_MODULE_V2,
            DeckSlotName.SLOT_B1,
        ),
    ],
)
def test_load_module_thermocycler_with_no_location(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_hardware_api: SyncHardwareAPI,
    requested_model: ModuleModel,
    engine_model: EngineModuleModel,
    subject: ProtocolCore,
    expected_slot: DeckSlotName,
) -> None:
    """It should issue a load module engine command with location at 7."""
    mock_hw_mod = decoy.mock(cls=AbstractModule)
    decoy.when(mock_hw_mod.device_info).then_return({"serial": "xyz789"})
    decoy.when(mock_sync_hardware_api.attached_modules).then_return([mock_hw_mod])
    decoy.when(mock_engine_client.state.config.robot_type).then_return("OT-3 Standard")

    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.LoadModuleParams(
                model=engine_model,
                location=DeckSlotLocation(slotName=expected_slot),
            )
        )
    ).then_return(
        commands.LoadModuleResult(
            moduleId="abc123",
            model=engine_model,
            serialNumber="xyz789",
        )
    )

    result = subject.load_module(
        model=requested_model,
        deck_slot=None,
        configuration=None,
    )

    decoy.verify(
        deck_conflict.check(
            engine_state=mock_engine_client.state,
            existing_labware_ids=[],
            existing_module_ids=[],
            existing_disposal_locations=[],
            new_module_id="abc123",
        )
    )

    assert isinstance(result, ThermocyclerModuleCore)
    assert result.module_id == "abc123"


@pytest.mark.parametrize(
    "requested_model",
    [
        HeaterShakerModuleModel.HEATER_SHAKER_V1,
        MagneticModuleModel.MAGNETIC_V1,
        MagneticModuleModel.MAGNETIC_V2,
        TemperatureModuleModel.TEMPERATURE_V1,
        TemperatureModuleModel.TEMPERATURE_V2,
        FlexStackerModuleModel.FLEX_STACKER_V1,
    ],
)
def test_load_module_no_location(
    requested_model: ModuleModel, subject: ProtocolCore
) -> None:
    """Should raise an InvalidModuleLocationError exception."""
    with pytest.raises(InvalidModuleLocationError):
        subject.load_module(model=requested_model, deck_slot=None, configuration=None)


@pytest.mark.parametrize("message", [None, "Hello, world!", ""])
def test_pause(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
    message: Optional[str],
) -> None:
    """It should issue a waitForResume command."""
    subject.pause(msg=message)
    decoy.verify(
        mock_engine_client.execute_command(cmd.WaitForResumeParams(message=message))
    )


@pytest.mark.parametrize("seconds", [0.0, -1.23, 1.23])
@pytest.mark.parametrize("message", [None, "Hello, world!", ""])
def test_delay(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
    seconds: float,
    message: Optional[str],
) -> None:
    """It should issue a waitForDuration command."""
    subject.delay(seconds=seconds, msg=message)
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.WaitForDurationParams(seconds=seconds, message=message)
        )
    )


def test_wait_for_tasks(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a waitForTasks command."""
    task1 = decoy.mock(cls=EngineTaskCore)
    task2 = decoy.mock(cls=EngineTaskCore)
    tasks = [task1, task2]
    task_ids = ["task-id-1", "task-id-2"]

    decoy.when(task1._id).then_return(task_ids[0])
    decoy.when(task2._id).then_return(task_ids[1])

    subject.wait_for_tasks(task_cores=tasks)
    decoy.verify(
        mock_engine_client.execute_command(cmd.WaitForTasksParams(task_ids=task_ids))
    )


def test_create_timer(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a createTimer command."""
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.CreateTimerParams(time=0.1)
        )
    ).then_return(cmd.CreateTimerResult(task_id="taskid", time=0.1))
    result = subject.create_timer(seconds=0.1)
    assert result._id == "taskid"
    assert result._engine_client == mock_engine_client


def test_comment(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should issue a comment command."""
    subject.comment("Hello, world!")
    decoy.verify(
        mock_engine_client.execute_command(cmd.CommentParams(message="Hello, world!"))
    )


def test_home(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should home all axes."""
    subject.home()
    decoy.verify(mock_engine_client.execute_command(cmd.HomeParams(axes=None)), times=1)


def test_is_simulating(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should return if simulating."""
    decoy.when(mock_engine_client.state.config.ignore_pause).then_return(True)
    assert subject.is_simulating()


def test_set_rail_lights(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should verify a call to sync client."""
    subject.set_rail_lights(on=True)
    decoy.verify(mock_engine_client.execute_command(cmd.SetRailLightsParams(on=True)))

    subject.set_rail_lights(on=False)
    decoy.verify(mock_engine_client.execute_command(cmd.SetRailLightsParams(on=False)))


def test_get_rail_lights(
    decoy: Decoy, mock_sync_hardware_api: SyncHardwareAPI, subject: ProtocolCore
) -> None:
    """It should get rails light state."""
    decoy.when(mock_sync_hardware_api.get_lights()).then_return({"rails": True})

    result = subject.get_rail_lights_on()
    assert result is True


def test_get_deck_definition(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should return the loaded deck definition from engine state."""
    deck_definition = cast(DeckDefinitionV5, {"schemaVersion": "5"})

    decoy.when(mock_engine_client.state.labware.get_deck_definition()).then_return(
        deck_definition
    )

    result = subject.get_deck_definition()

    assert result == deck_definition


def test_get_labware_on_module(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should get the item on top of a given module."""
    mock_module_core = decoy.mock(cls=ModuleCore)
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_module_core.module_id).then_return("abc")
    decoy.when(mock_engine_client.state.labware.get_id_by_module("abc")).then_return(
        "123"
    )

    subject._labware_cores_by_id["123"] = mock_labware_core

    assert subject.get_labware_on_module(mock_module_core) == mock_labware_core


def test_get_labware_on_module_returns_none(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should return none if there is no item on top of the module."""
    mock_module_core = decoy.mock(cls=ModuleCore)

    decoy.when(mock_module_core.module_id).then_return("abc")
    decoy.when(mock_engine_client.state.labware.get_id_by_module("abc")).then_raise(
        LabwareNotLoadedOnModuleError("whoops")
    )

    assert subject.get_labware_on_module(mock_module_core) is None


def test_get_labware_on_labware(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should get the item on top of a given labware."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    mock_other_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_labware_core.labware_id).then_return("abc")
    decoy.when(mock_engine_client.state.labware.get_id_by_labware("abc")).then_return(
        "123"
    )

    subject._labware_cores_by_id["123"] = mock_other_labware_core

    assert subject.get_labware_on_labware(mock_labware_core) == mock_other_labware_core


def test_get_labware_on_labware_returns_none(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should return none if there is no item on top of the labware."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_labware_core.labware_id).then_return("abc")
    decoy.when(mock_engine_client.state.labware.get_id_by_labware("abc")).then_raise(
        LabwareNotLoadedOnLabwareError("oops")
    )

    assert subject.get_labware_on_labware(mock_labware_core) is None


def test_get_slot_center(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should return a slot center from engine state."""
    decoy.when(
        mock_engine_client.state.addressable_areas.get_addressable_area_center(
            DeckSlotName.SLOT_2.id
        )
    ).then_return(Point(1, 2, 3))

    result = subject.get_slot_center(DeckSlotName.SLOT_2)

    assert result == Point(1, 2, 3)


def test_get_highest_z(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ProtocolCore
) -> None:
    """It should return a slot center from engine state."""
    decoy.when(
        mock_engine_client.state.geometry.get_all_obstacle_highest_z()
    ).then_return(9001)

    result = subject.get_highest_z()

    assert result == 9001


def test_add_liquid(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should return the created liquid."""
    liquid = PE_Liquid.model_construct(
        id="water-id",
        displayName="water",
        description="water desc",
        displayColor=HexColor("#fff"),
    )

    expected_result = Liquid(
        _id="water-id",
        name="water",
        description="water desc",
        display_color="#fff",
    )

    decoy.when(
        mock_engine_client.add_liquid(
            name="water", color="#fff", description="water desc"
        )
    ).then_return(liquid)

    result = subject.define_liquid(
        name="water", description="water desc", display_color="#fff"
    )

    assert result == expected_result


def test_define_liquid_class(
    decoy: Decoy,
    subject: ProtocolCore,
    minimal_liquid_class_def1: LiquidClassSchemaV1,
    minimal_liquid_class_def2: LiquidClassSchemaV1,
) -> None:
    """It should create a LiquidClass and cache the definition."""
    expected_liquid_class = LiquidClass(
        _name="water1", _display_name="water 1", _by_pipette_setting={}
    )
    decoy.when(liquid_classes.load_definition("water", version=123)).then_return(
        minimal_liquid_class_def1
    )
    assert subject.get_liquid_class("water", 123) == expected_liquid_class

    # Test that different version number works
    decoy.when(liquid_classes.load_definition("water", version=456)).then_return(
        minimal_liquid_class_def2
    )
    different_liquid_class = subject.get_liquid_class("water", 456)
    assert different_liquid_class.name == "water2"
    assert different_liquid_class.display_name == "water 2"

    # Test that definition caching works
    decoy.when(liquid_classes.load_definition("water", version=123)).then_return(
        minimal_liquid_class_def2
    )
    assert subject.get_liquid_class("water", 123) == expected_liquid_class


def test_define_liquid_class_without_version_provided(
    decoy: Decoy,
    subject: ProtocolCore,
    minimal_liquid_class_def1: LiquidClassSchemaV1,
    minimal_liquid_class_def2: LiquidClassSchemaV1,
) -> None:
    """It should create a LiquidClass with the most recent version and cache the definition."""
    expected_liquid_class = LiquidClass(
        _name="water1", _display_name="water 1", _by_pipette_setting={}
    )
    decoy.when(
        _default_liquid_class_versions.get_liquid_class_version(
            subject.api_version, "water"
        )
    ).then_return(987)
    decoy.when(liquid_classes.load_definition("water", version=987)).then_return(
        minimal_liquid_class_def1
    )

    assert subject.get_liquid_class("water", version=None) == expected_liquid_class

    # Test that specified version number works too
    decoy.when(liquid_classes.load_definition("water", version=654)).then_return(
        minimal_liquid_class_def2
    )
    different_liquid_class = subject.get_liquid_class("water", 654)
    assert different_liquid_class.name == "water2"
    assert different_liquid_class.display_name == "water 2"

    # Test that definition caching works
    decoy.when(liquid_classes.load_definition("water", version=987)).then_return(
        minimal_liquid_class_def2
    )
    assert subject.get_liquid_class("water", version=None) == expected_liquid_class


def test_get_labware_location_deck_slot(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should return the labware location as a deck slot string."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_labware_core.labware_id).then_return("abc")

    decoy.when(mock_engine_client.state.labware.get_location("abc")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )
    decoy.when(mock_engine_client.state.config.robot_type).then_return("OT-2 Standard")
    decoy.when(
        validation.internal_slot_to_public_string(DeckSlotName.SLOT_1, "OT-2 Standard")
    ).then_return("777")

    assert subject.get_labware_location(mock_labware_core) == "777"


def test_get_labware_location_module(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should return the labware location as a module."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_labware_core.labware_id).then_return("abc")

    decoy.when(mock_engine_client.state.labware.get_location("abc")).then_return(
        ModuleLocation(moduleId="123")
    )

    mock_module_core = decoy.mock(cls=ModuleCore)
    subject._module_cores_by_id["123"] = mock_module_core

    assert subject.get_labware_location(mock_labware_core) == mock_module_core


def test_get_labware_location_labware(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should return the labware location as a labware."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_labware_core.labware_id).then_return("abc")

    decoy.when(mock_engine_client.state.labware.get_location("abc")).then_return(
        OnLabwareLocation(labwareId="123")
    )

    mock_parent_labware_core = decoy.mock(cls=LabwareCore)
    subject._labware_cores_by_id["123"] = mock_parent_labware_core

    assert subject.get_labware_location(mock_labware_core) == mock_parent_labware_core


def test_get_labware_location_off_deck(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ProtocolCore,
) -> None:
    """It should return the labware location as None to represent an off deck labware."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_labware_core.labware_id).then_return("abc")

    decoy.when(mock_engine_client.state.labware.get_location("abc")).then_return(
        "offDeck"
    )

    assert subject.get_labware_location(mock_labware_core) is OFF_DECK


def test_capture_image_with_run_specific_defaults(
    decoy: Decoy,
    subject: ProtocolCore,
    mock_engine_client: EngineClient,
) -> None:
    """It should handle the core execution for capture image using the run specific image settings."""
    decoy.when(mock_engine_client.state.camera.get_resolution()).then_return((10, 20))
    decoy.when(mock_engine_client.state.camera.get_zoom()).then_return(1.5)
    decoy.when(mock_engine_client.state.camera.get_contrast()).then_return(25)
    decoy.when(mock_engine_client.state.camera.get_brightness()).then_return(75)
    decoy.when(mock_engine_client.state.camera.get_saturation()).then_return(99)
    subject.capture_image()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.CaptureImageParams(
                fileName=None,
                resolution=(10, 20),
                zoom=1.5,
                contrast=25,
                brightness=75,
                saturation=99,
            )
        )
    )



"""Arguments needed to test inner geometry.

Each labware has 2 nominal volumes calculated in solidworks.
- One is a nominal bottom volume, calculated some set distance from the bottom of the inside of the well.
- The other is a nominal top volume, calculated some set distance from the top of the inside of the well.
"""

INNER_WELL_GEOMETRY_TEST_PARAMS = [
    [
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical",
        "conicalWell15mL",
        16.69,
        15546.930,
        3.0,
        5.0,
    ],
    [
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical",
        "conicalWell50mL",
        111.2,
        56110.279,
        3.0,
        5.0,
    ],
    [
        "opentrons_24_tuberack_nest_2ml_screwcap",
        "conicalWell",
        64.4,
        2035.5,
        2.96,
        4.22,
    ],
    [
        "opentrons_24_tuberack_nest_1.5ml_screwcap",
        "conicalWell",
        18.8,
        1693.0,
        2.94,
        4.02,
    ],
    ["nest_1_reservoir_290ml", "cuboidalWell", 16570.12, 271690.52, 3.0, 3.0],
    [
        "opentrons_24_tuberack_nest_2ml_snapcap",
        "conicalWell",
        67.323,
        2077.6,
        2.946,
        4.06,
    ],
    ["nest_96_wellplate_2ml_deep", "cuboidalWell", 114.396, 1992.407, 2.93, 4.03],
    [
        "opentrons_24_tuberack_nest_1.5ml_snapcap",
        "conicalWell",
        27.8,
        1682.29,
        3.0,
        3.0,
    ],
    ["nest_12_reservoir_15ml", "cuboidalWell", 1178.77, 13236.1, 2.93, 3.0],
    ["nest_1_reservoir_195ml", "cuboidalWell", 14034.19, 172301.86, 3.0, 3.0],
    [
        "opentrons_24_tuberack_nest_0.5ml_screwcap",
        "conicalWell",
        21.226,
        769.19,
        3.15,
        3.35,
    ],
    [
        "opentrons_96_wellplate_200ul_pcr_full_skirt",
        "conicalWell",
        13.8,
        145.2,
        2.93,
        3.21,
    ],
    ["nest_96_wellplate_100ul_pcr_full_skirt", "conicalWell", 14.99, 145.79, 2.93, 3.2],
    ["nest_96_wellplate_200ul_flat", "conicalWell", 93.1, 251.2, 2.9, 3.24],
    [
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical",
        "50mlconicalWell",
        163.860,
        57720.510,
        3.0,
        3.0,
    ],
    [
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical",
        "15mlconicalWell",
        40.830,
        15956.6,
        3.0,
        3.0,
    ],
    ["usascientific_12_reservoir_22ml", "cuboidalWell", 66.17 * 8, 19797.38, 3.0, 6.0],
    [
        "thermoscientificnunc_96_wellplate_2000ul",
        "conicalWell",
        71.1,
        1768.02,
        2.94,
        3.0,
    ],
    ["usascientific_96_wellplate_2.4ml_deep", "cuboidalWell", 72.22, 2241.36, 3.0, 3.0],
    ["agilent_1_reservoir_290ml", "cuboidalWell", 15652.89, 268813.81, 3.0, 3.0],
    [
        "opentrons_24_tuberack_eppendorf_1.5ml_safelock_snapcap",
        "conicalWell",
        25.75,
        1576.14,
        3.0,
        3.0,
    ],
    [
        "thermoscientificnunc_96_wellplate_1300ul",
        "conicalWell",
        71.1,
        1117.0,
        2.94,
        3.71,
    ],
    ["corning_12_wellplate_6.9ml_flat", "conicalWell", 1118.1, 5468.2, 2.9, 3.46],
    ["corning_24_wellplate_3.4ml_flat", "conicalWell", 559.9, 2759.3, 2.9, 3.46],
    ["corning_6_wellplate_16.8ml_flat", "conicalWell", 2767.7, 13443.1, 2.9, 3.47],
    ["corning_48_wellplate_1.6ml_flat", "conicalWell", 260.0, 1283.2, 2.9, 3.46],
    ["biorad_96_wellplate_200ul_pcr", "conicalWell", 17.3, 155.9, 2.93, 3.23],
    ["axygen_1_reservoir_90ml", "cuboidalWell", 14909.08, 72854.8, 2.0, 2.68],
    ["corning_384_wellplate_112ul_flat", "flatWell", 22.43, 77.39, 2.88, 3.0],
    ["corning_96_wellplate_360ul_flat", "conicalWell", 94.0, 248.6, 2.9, 3.24],
    ["biorad_384_wellplate_50ul", "conicalWell", 7.7, 27.76, 3.0, 3.0],
    [
        "appliedbiosystemsmicroamp_384_wellplate_40ul",
        "conicalWell",
        7.2,
        25.3,
        2.94,
        3.11,
    ],
    [
        "opentrons_24_tuberack_eppendorf_2ml_safelock_snapcap",
        "conicalWell",
        60.94,
        2163.98,
        3.0,
        3.0,
    ],
    [
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical",
        "conicalWell15mL",
        16.69,
        15546.93,
        3.0,
        5.0,
    ],
    [
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical",
        "conicalWell50mL",
        111.2,
        56110.279,
        3.0,
        5.0,
    ],
    [
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical",
        "15mlconicalWell",
        40.83,
        15956.6,
        3.0,
        3.0,
    ],
    [
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical",
        "50mlconicalWell",
        163.86,
        57720.51,
        3.0,
        3.0,
    ],
]



from pathlib import Path
from typing import TYPE_CHECKING, Type, Union

import pytest

from opentrons_shared_data.robot.types import RobotType

from opentrons.config import robot_configs
from opentrons.hardware_control import API, simulator_setup
from opentrons.hardware_control.modules import MagDeck, TempDeck, Thermocycler
from opentrons.hardware_control.types import OT3Mount
from opentrons.types import Mount

if TYPE_CHECKING:
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture
def setup_klass(robot_model: RobotType) -> Type[simulator_setup.SimulatorSetup]:
    """Get the appropriate setup class for the machine type."""
    if robot_model == "OT-2 Standard":
        return simulator_setup.OT2SimulatorSetup
    else:
        return simulator_setup.OT3SimulatorSetup


@pytest.fixture
def simulator_type(robot_model: RobotType) -> Union[Type[API], Type["OT3API"]]:
    """Get the appropriate simulated hardware controller instance type."""
    if robot_model == "OT-2 Standard":
        return API
    else:
        from opentrons.hardware_control.ot3api import OT3API

        return OT3API


async def assure_simulator_type(
    setup_klass: Type[simulator_setup.SimulatorSetup],
    simulator_type: Union[Type[API], Type["OT3API"]],
) -> None:
    """It should create the appropriate kind of direct simulator."""
    simulator = await simulator_setup.create_simulator(setup_klass())
    assert isinstance(simulator, simulator_type)


async def assure_thread_manager_type(
    setup_klass: Type[simulator_setup.SimulatorSetup],
    simulator_type: Union[Type[API], Type["OT3API"]],
) -> None:
    """It should create the appropriate kind of thread manager simulator."""
    manager = await simulator_setup.create_simulator_thread_manager(setup_klass())
    assert isinstance(object.__getattribute__(manager, "managed_obj"), simulator_type)


async def test_with_magdeck(setup_klass: Type[simulator_setup.SimulatorSetup]) -> None:
    """It should work to build a magdeck."""
    setup = setup_klass(
        attached_modules={
            "magdeck": [
                simulator_setup.ModuleItem(
                    model="magneticModuleV1",
                    serial_number="123",
                    calls=[simulator_setup.ModuleCall("engage", kwargs={"height": 3})],
                ),
                simulator_setup.ModuleItem(
                    model="magneticModuleV2",
                    serial_number="1234",
                    calls=[simulator_setup.ModuleCall("engage", kwargs={"height": 5})],
                ),
            ]
        }
    )
    simulator = await simulator_setup.create_simulator(setup)

    assert isinstance(simulator.attached_modules[0], MagDeck)
    assert simulator.attached_modules[0].model() == "magneticModuleV1"
    assert simulator.attached_modules[0].live_data == {
        "data": {"engaged": True, "height": 3},
        "status": "engaged",
    }
    assert simulator.attached_modules[0].device_info["serial"] == "123"
    assert simulator.attached_modules[1].model() == "magneticModuleV2"
    assert simulator.attached_modules[1].live_data == {
        "data": {"engaged": True, "height": 5},
        "status": "engaged",
    }
    assert simulator.attached_modules[1].device_info["serial"] == "1234"


async def test_with_thermocycler(
    setup_klass: Type[simulator_setup.SimulatorSetup],
) -> None:
    """It should work to build a thermocycler."""
    setup = setup_klass(
        attached_modules={
            "thermocycler": [
                simulator_setup.ModuleItem(
                    model="thermocyclerModuleV2",
                    serial_number="123",
                    calls=[
                        simulator_setup.ModuleCall(
                            "set_temperature",
                            kwargs={
                                "temperature": 3,
                                "hold_time_seconds": 1,
                                "hold_time_minutes": 2,
                                "volume": 5,
                            },
                        )
                    ],
                )
            ]
        }
    )
    simulator = await simulator_setup.create_simulator(setup)

    assert isinstance(simulator.attached_modules[0], Thermocycler)
    assert simulator.attached_modules[0].model() == "thermocyclerModuleV2"
    assert simulator.attached_modules[0].live_data == {
        "data": {
            "currentCycleIndex": None,
            "currentStepIndex": None,
            "currentTemp": 3,
            "holdTime": 0,
            "lid": "open",
            "lidTarget": None,
            "lidTemp": 23,
            "lidTempStatus": "idle",
            "rampRate": None,
            "targetTemp": 3,
            "totalCycleCount": None,
            "totalStepCount": None,
        },
        "status": "holding at target",
    }
    assert simulator.attached_modules[0].device_info["serial"] == "123"


async def test_with_tempdeck(setup_klass: Type[simulator_setup.SimulatorSetup]) -> None:
    """It should work to build a tempdeck."""
    setup = setup_klass(
        attached_modules={
            "tempdeck": [
                simulator_setup.ModuleItem(
                    model="temperatureModuleV2",
                    serial_number="123",
                    calls=[
                        simulator_setup.ModuleCall(
                            "start_set_temperature", kwargs={"celsius": 23}
                        ),
                        simulator_setup.ModuleCall(
                            "await_temperature", kwargs={"awaiting_temperature": None}
                        ),
                    ],
                )
            ]
        }
    )
    simulator = await simulator_setup.create_simulator(setup)

    assert isinstance(simulator.attached_modules[0], TempDeck)
    assert simulator.attached_modules[0].model() == "temperatureModuleV2"
    assert simulator.attached_modules[0].live_data == {
        "data": {"currentTemp": 23, "targetTemp": 23},
        "status": "holding at target",
    }
    assert simulator.attached_modules[0].device_info["serial"] == "123"


def test_persistence_ot2(tmpdir: str) -> None:
    sim = simulator_setup.OT2SimulatorSetup(
        attached_instruments={
            Mount.LEFT: {"id": "an id"},
            Mount.RIGHT: {"id": "some id"},
        },
        attached_modules={
            "magdeck": [
                simulator_setup.ModuleItem(
                    model="magneticModuleV1",
                    serial_number="111",
                    calls=[simulator_setup.ModuleCall("engage", kwargs={"height": 3})],
                )
            ],
            "tempdeck": [
                simulator_setup.ModuleItem(
                    model="temperatureModuleV2",
                    serial_number="111",
                    calls=[
                        simulator_setup.ModuleCall(
                            "set_temperature", kwargs={"celsius": 23}
                        ),
                        simulator_setup.ModuleCall(
                            "set_temperature", kwargs={"celsius": 24}
                        ),
                    ],
                )
            ],
        },
        config=robot_configs.build_config_ot2({}),
    )
    file = Path(tmpdir) / "sim_setup.json"
    simulator_setup.save_simulator_setup(sim, file)
    test_sim = simulator_setup.load_simulator_setup(file)

    assert test_sim == sim


def test_persistence_ot3(tmpdir: str) -> None:
    sim = simulator_setup.OT3SimulatorSetup(
        attached_instruments={
            OT3Mount.LEFT: {"id": "an id"},
            OT3Mount.RIGHT: {"id": "some id"},
            OT3Mount.GRIPPER: {"id": "some-other-id"},
        },
        attached_modules={
            "magdeck": [
                simulator_setup.ModuleItem(
                    model="magneticModuleV2",
                    serial_number="mag-1",
                    calls=[
                        simulator_setup.ModuleCall(
                            function_name="engage",
                            kwargs={"height": 3},
                        )
                    ],
                )
            ],
            "tempdeck": [
                simulator_setup.ModuleItem(
                    model="temperatureModuleV2",
                    serial_number="temp-1",
                    calls=[
                        simulator_setup.ModuleCall(
                            function_name="set_temperature",
                            kwargs={"celsius": 23},
                        ),
                        simulator_setup.ModuleCall(
                            function_name="set_temperature",
                            kwargs={"celsius": 24},
                        ),
                    ],
                ),
                simulator_setup.ModuleItem(
                    model="temperatureModuleV1",
                    serial_number="temp-2",
                    calls=[
                        simulator_setup.ModuleCall(
                            function_name="set_temperature",
                            kwargs={"celsius": 23},
                        ),
                        simulator_setup.ModuleCall(
                            function_name="set_temperature",
                            kwargs={"celsius": 24},
                        ),
                    ],
                ),
            ],
        },
        config=robot_configs.build_config_ot3({}),
    )
    file = Path(tmpdir) / "sim_setup.json"
    simulator_setup.save_simulator_setup(sim, file)
    test_sim = simulator_setup.load_simulator_setup(file)

    assert test_sim == sim



"""Tests for Protocol API magnetic module contexts."""

import pytest
from decoy import Decoy, matchers

from opentrons.hardware_control.modules import MagneticStatus
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import MAX_SUPPORTED_VERSION, MagneticModuleContext
from opentrons.protocol_api.core.common import MagneticModuleCore, ProtocolCore
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import UnsupportedAPIError


@pytest.fixture
def mock_core(decoy: Decoy) -> MagneticModuleCore:
    """Get a mock module implementation core."""
    return decoy.mock(cls=MagneticModuleCore)


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol implementation core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock LoadedCoreMap."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def mock_broker(decoy: Decoy) -> LegacyBroker:
    """Get a mock command message broker."""
    return decoy.mock(cls=LegacyBroker)


@pytest.fixture
def api_version() -> APIVersion:
    """Get an API version to apply to the interface."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def subject(
    api_version: APIVersion,
    mock_core: MagneticModuleCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_broker: LegacyBroker,
) -> MagneticModuleContext:
    """Get a magnetic module context with its dependencies mocked out."""
    return MagneticModuleContext(
        core=mock_core,
        protocol_core=mock_protocol_core,
        core_map=mock_core_map,
        broker=mock_broker,
        api_version=api_version,
    )


def test_disengage(
    decoy: Decoy,
    mock_core: MagneticModuleCore,
    mock_broker: LegacyBroker,
    subject: MagneticModuleContext,
) -> None:
    """It should disengage magnets via the core."""
    subject.disengage()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "before", "name": "command.MAGDECK_DISENGAGE"}),
        ),
        mock_core.disengage(),
        mock_broker.publish("command", matchers.DictMatching({"$": "after"})),
    )


def test_get_status(
    decoy: Decoy, mock_core: MagneticModuleCore, subject: MagneticModuleContext
) -> None:
    """It should report the status from the core."""
    decoy.when(mock_core.get_status()).then_return(MagneticStatus.DISENGAGED)

    result = subject.status

    assert result == "disengaged"


@pytest.mark.parametrize("api_version", [APIVersion(2, 13)])
def test_engage_height_from_home_succeeds_on_low_version(
    decoy: Decoy,
    mock_broker: LegacyBroker,
    mock_core: MagneticModuleCore,
    subject: MagneticModuleContext,
) -> None:
    """It should engage if given a raw motor height and the apiLevel is low."""
    subject.engage(height=42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "before", "name": "command.MAGDECK_ENGAGE"}),
        ),
        mock_core.engage(height_from_home=42.0),
        mock_broker.publish("command", matchers.DictMatching({"$": "after"})),
    )


# TODO(mm, 2023-02-09): Add MAX_SUPPORTED_VERSION when it's >=2.14.
@pytest.mark.parametrize("api_version", [APIVersion(2, 14)])
def test_engage_height_from_home_raises_on_high_version(
    decoy: Decoy,
    mock_core: MagneticModuleCore,
    subject: MagneticModuleContext,
) -> None:
    """It should error if given a raw motor height and the apiLevel is high."""
    with pytest.raises(UnsupportedAPIError):
        subject.engage(height=42.0)
    with pytest.raises(UnsupportedAPIError):
        subject.engage(42.0)


@pytest.mark.parametrize("api_version", [APIVersion(2, 14)])
def test_calibrate_raises_on_high_version(
    decoy: Decoy,
    mock_core: MagneticModuleCore,
    subject: MagneticModuleContext,
) -> None:
    """It should raise a deprecation error."""
    with pytest.raises(UnsupportedAPIError):
        subject.calibrate()


def test_engage_height_from_base(
    decoy: Decoy,
    mock_broker: LegacyBroker,
    mock_core: MagneticModuleCore,
    subject: MagneticModuleContext,
) -> None:
    """It should engage if given a height from the base."""
    subject.engage(height_from_base=42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "before", "name": "command.MAGDECK_ENGAGE"}),
        ),
        mock_core.engage(height_from_base=42.0),
        mock_broker.publish("command", matchers.DictMatching({"$": "after"})),
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 3)])
def test_engage_offset_from_default(
    decoy: Decoy,
    mock_broker: LegacyBroker,
    mock_core: MagneticModuleCore,
    subject: MagneticModuleContext,
) -> None:
    """It should engage from a labware's default engage height."""
    subject.engage(offset=42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "before", "name": "command.MAGDECK_ENGAGE"}),
        ),
        mock_core.engage_to_labware(offset=42.0, preserve_half_mm=False),
        mock_broker.publish("command", matchers.DictMatching({"$": "after"})),
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 2)])
def test_engage_offset_from_default_low_version(
    decoy: Decoy,
    mock_core: MagneticModuleCore,
    subject: MagneticModuleContext,
) -> None:
    """It should preserve pre-2.3 buggy labware engage height behavior."""
    subject.engage(offset=42.0)

    decoy.verify(
        mock_core.engage_to_labware(offset=42.0, preserve_half_mm=True),
        times=1,
    )


def test_serial_number(
    decoy: Decoy,
    mock_core: MagneticModuleCore,
    subject: MagneticModuleContext,
) -> None:
    """It should get the module's unique serial number."""
    decoy.when(mock_core.get_serial_number()).then_return("abc-123")
    result = subject.serial_number
    assert result == "abc-123"



import json
from textwrap import dedent
from typing import Any, Callable, Literal, Optional, Union

import pytest

from opentrons_shared_data.robot.types import RobotType

from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.parse import (
    API_VERSION_FOR_JSON_V5_AND_BELOW,
    MAX_SUPPORTED_JSON_SCHEMA_VERSION,
    JSONSchemaVersionTooNewError,
    PythonParseMode,
    _get_protocol_schema_version,
    parse,
    validate_json,
)
from opentrons.protocols.types import (
    ApiDeprecationError,
    JsonProtocol,
    MalformedPythonProtocolError,
    Protocol,
    PythonProtocol,
    PythonProtocolMetadata,
)

parse_version_cases = [
    # No explicitly-declared apiLevel (infer heuristically from APIv1 imports):
    (
        """
        from opentrons import types, containers
        """,
        APIVersion(1, 0),
    ),
    (
        """
        from opentrons import instruments

        p = instruments.P10_Single(mount='right')
        """,
        APIVersion(1, 0),
    ),
    (
        """
        from opentrons import labware, instruments

        p = instruments.P10_Single(mount='right')
        """,
        APIVersion(1, 0),
    ),
    (
        """
        from opentrons import types, instruments

        p = instruments.P10_Single(mount='right')
        """,
        APIVersion(1, 0),
    ),
    (
        """
        from opentrons import instruments as instr

        p = instr.P10_Single(mount='right')
        """,
        APIVersion(1, 0),
    ),
    (
        """
        import opentrons.instruments

        p = instruments.P10_Single(mount='right')
        """,
        APIVersion(1, 0),
    ),
    # Explicitly-declared APIv1:
    (
        """
        from opentrons import instruments

        metadata = {
          'apiLevel': '1'
          }

        p = instruments.P10_Single(mount='right')
        """,
        APIVersion(1, 0),
    ),
    (
        # Explicitly-declared APIv1 despite having an APIv2-style run(ctx) function.
        """
        from opentrons import types

        metadata = {
          'apiLevel': '1'
          }

        def run(ctx):
            right = ctx.load_instrument('p300_single', types.Mount.RIGHT)
        """,
        APIVersion(1, 0),
    ),
    (
        # apiLevel as major.minor, not just major.
        # APIv1 didn't have minor apiLevel versions, so nobody would have actually written this in
        # the APIv1 days, but we should make sure we don't choke on it.
        """
        metadata = {
          'apiLevel': '1.0'
        }

        p = instruments.P10_Single(mount='right')
        """,
        APIVersion(1, 0),
    ),
    (
        # apiLevel as major.minor, not just major.
        # APIv1 didn't have minor apiLevel versions, so nobody would have actually written this in
        # the APIv1 days, but we should make sure we don't choke on it.
        """
        metadata = {
          'apiLevel': '1.2'
        }

        p = instruments.P10_Single(mount='right')
        """,
        APIVersion(1, 2),
    ),
    # Explicitly-declared APIv2 or above:
    (
        """
        from opentrons import types

        metadata = {
            'apiLevel': '2.0'
        }

        def run(ctx):
            right = ctx.load_instrument('p300_single', types.Mount.RIGHT)
        """,
        APIVersion(2, 0),
    ),
    (
        """
        from opentrons import types

        metadata = {
          'apiLevel': '2.0'
          }

        def run(ctx):
            right = ctx.load_instrument('p300_single', types.Mount.RIGHT)
        """,
        APIVersion(2, 0),
    ),
    (
        """
        metadata = {'apiLevel': '2.6'}
        def run(ctx): pass
        """,
        APIVersion(2, 6),
    ),
    (
        """
        metadata = {'apiLevel': '10.23123151'}
        def run(ctx): pass
        """,
        APIVersion(10, 23123151),
    ),
    # Explicitly-declared apiLevel with various cases of it being in metadata or requirements:
    (
        """
        requirements = {"apiLevel": "123.456"}
        def run(ctx): pass
        """,
        APIVersion(123, 456),
    ),
    (
        """
        metadata = {}
        requirements = {"apiLevel": "123.456"}
        def run(ctx): pass
        """,
        APIVersion(123, 456),
    ),
    (
        """
        metadata = {"apiLevel": "123.456"}
        def run(ctx): pass
        """,
        APIVersion(123, 456),
    ),
    (
        """
        metadata = {"apiLevel": "123.456"}
        requirements = {}
        def run(ctx): pass
        """,
        APIVersion(123, 456),
    ),
]


@pytest.mark.parametrize("proto,version", parse_version_cases)
def test_parse_get_version(proto: str, version: APIVersion) -> None:
    proto = dedent(proto)
    if version < APIVersion(2, 0):
        with pytest.raises(ApiDeprecationError):
            parse(proto)
    else:
        parsed = parse(proto)
        assert parsed.api_level == version


@pytest.mark.parametrize(
    ("protocol_source", "expected_message"),
    [
        (
            """
            metadata = {"apiLevel": "2"}
            def run(cxt): pass
            """,
            "incorrectly formatted. It should be major.minor",
        ),
        (
            """
            metadata = {"apiLevel": "2.0.0"}
            def run(cxt): pass
            """,
            "incorrectly formatted. It should be major.minor",
        ),
        (
            """
            metadata = {"apiLevel": "asda"}
            def run(cxt): pass
            """,
            "incorrectly formatted. It should be major.minor",
        ),
    ],
)
def test_version_from_static_python_info_invalid(
    protocol_source: str, expected_message: str
) -> None:
    with pytest.raises(MalformedPythonProtocolError, match=expected_message):
        parse(dedent(protocol_source), "protocol.py")


def test_get_protocol_schema_version() -> None:
    assert _get_protocol_schema_version({"protocol-schema": "1.0.0"}) == 1
    assert _get_protocol_schema_version({"protocol-schema": "2.0.0"}) == 2
    assert _get_protocol_schema_version({"schemaVersion": 123}) == 123

    # schemaVersion has precedence over legacy 'protocol-schema'
    assert (
        _get_protocol_schema_version({"protocol-schema": "2.0.0", "schemaVersion": 123})
        == 123
    )

    with pytest.raises(RuntimeError):
        _get_protocol_schema_version({"schemaVersion": None})
    with pytest.raises(RuntimeError):
        _get_protocol_schema_version({})
    with pytest.raises(RuntimeError):
        _get_protocol_schema_version({"protocol-schema": "1.2.3"})


def test_validate_json(
    get_json_protocol_fixture: Callable[..., Any],
    get_labware_fixture: Callable[..., Any],
) -> None:
    # valid data that has no schema should fail
    with pytest.raises(RuntimeError, match="deprecated"):
        validate_json({"protocol-schema": "1.0.0"})
    with pytest.raises(JSONSchemaVersionTooNewError):
        validate_json({"schemaVersion": str(MAX_SUPPORTED_JSON_SCHEMA_VERSION + 1)})
    labware = get_labware_fixture("fixture_12_trough_v2")
    with pytest.raises(RuntimeError, match="labware"):
        validate_json(labware)
    with pytest.raises(RuntimeError, match="corrupted"):
        validate_json({"schemaVersion": "3"})

    v3 = get_json_protocol_fixture("3", "testAllAtomicSingleV3")
    assert validate_json(v3)[0] == 3

    v4 = get_json_protocol_fixture("4", "testModulesProtocol")
    assert validate_json(v4)[0] == 4


@pytest.mark.parametrize(
    (
        "protocol_source",
        "expected_metadata",
        "expected_api_level",
        "expected_robot_type",
    ),
    [
        (
            # Basic APIv2 test with a bunch of stuff in metadata.
            """
            from opentrons import protocol_api, types

            metadata = {
                "protocolName": "Testosaur",
                "author": "Opentrons <engineering@opentrons.com>",
                "description": 'A variant on "Dinosaur" for testing',
                "source": "Opentrons Repository",
                "apiLevel": "2.0",
            }


            def run(ctx: protocol_api.ProtocolContext) -> None:
                ctx.home()
                tr = ctx.load_labware("opentrons_96_tiprack_1000ul", 1)
                right = ctx.load_instrument("p1000_single", types.Mount.RIGHT, [tr])
                lw = ctx.load_labware("corning_96_wellplate_360ul_flat", 2)
                right.pick_up_tip()
                right.aspirate(100, lw.wells()[0].bottom())
                right.dispense(100, lw.wells()[1].bottom())
                right.drop_tip(tr.wells()[-1].top())
            """,
            {
                "protocolName": "Testosaur",
                "author": "Opentrons <engineering@opentrons.com>",
                "description": 'A variant on "Dinosaur" for testing',
                "source": "Opentrons Repository",
                "apiLevel": "2.0",
            },
            APIVersion(2, 0),
            "OT-2 Standard",
        ),
        (
            # Both metadata and requirements, intermixed with a bunch of other statements.
            """
            this = 0
            that = 1
            metadata = {
                'mk1': 'mv1',
                'mk2': 'mv2',
                'apiLevel': '2.123'
            }
            print('wat?')
            def run(cxt): pass
            requirements = {
                'robotType': 'Flex'
            }
            metadata['hello'] = 'moon'
            fakedata['what?'] = 'ham'
            """,
            {
                "mk1": "mv1",
                "mk2": "mv2",
                "apiLevel": "2.123",
            },
            APIVersion(2, 123),
            "OT-3 Standard",
        ),
        (
            # Later assignments to metadata should override earlier ones.
            # TODO(mm, 2022-10-24): We're covering this with this test to retain current behavior,
            # but we should reconsider whether we actually want this behavior.
            # Protocols probably shouldn't do this. Note that metadata["k"] = "v" is
            # unsupported.
            """
            metadata = {'mk1': 'mv1'}
            metadata = {'mk2': 'mv2'}
            requirements = {'apiLevel': '2.0'}
            requirements = {'apiLevel': '2.1'}
            def run(cxt): pass
            """,
            {"mk2": "mv2"},
            APIVersion(2, 1),
            "OT-2 Standard",
        ),
        (
            # Explicitly-specified robotType.
            """
            requirements = {"apiLevel": "2.15", "robotType": "OT-2"}
            def run(ctx): pass
            """,
            None,
            APIVersion(2, 15),
            "OT-2 Standard",
        ),
        (
            # Explicitly-specified robotType.
            """
            requirements = {"apiLevel": "2.15", "robotType": "Flex"}
            def run(ctx): pass
            """,
            None,
            APIVersion(2, 15),
            "OT-3 Standard",
        ),
        (
            # Explicitly-specified robotType.
            """
            requirements = {"apiLevel": "2.15", "robotType": "OT-3"}
            def run(ctx): pass
            """,
            None,
            APIVersion(2, 15),
            "OT-3 Standard",
        ),
    ],
)
@pytest.mark.parametrize("protocol_text_kind", ["str", "bytes"])
@pytest.mark.parametrize("filename", ["protocol.py", None])
def test_parse_python_details(
    protocol_source: str,
    protocol_text_kind: Literal["str", "bytes"],
    filename: Optional[str],
    expected_api_level: APIVersion,
    expected_robot_type: RobotType,
    expected_metadata: PythonProtocolMetadata,
) -> None:
    protocol_source = dedent(protocol_source)

    if protocol_text_kind == "bytes":
        text: Union[bytes, str] = protocol_source.encode("utf-8")
    else:
        text = protocol_source

    parsed = parse(text, filename)

    assert isinstance(parsed, PythonProtocol)
    assert parsed.text == text
    if protocol_text_kind == "str":
        assert isinstance(parsed.text, str)
    else:
        assert isinstance(parsed.text, bytes)

    assert parsed.filename == filename
    assert parsed.contents.co_filename == (
        filename if filename is not None else "<protocol>"
    )

    assert parsed.api_level == expected_api_level
    assert expected_robot_type == expected_robot_type
    assert parsed.metadata == expected_metadata
    assert parsed.contents == compile(
        protocol_source,
        filename="<Python ignores this filename in this comparison>",
        mode="exec",
    )


@pytest.mark.parametrize(
    ("fixture_version", "fixture_name"),
    [
        ("3", "simple"),
        ("3", "testAllAtomicSingleV3"),
    ],
)
@pytest.mark.parametrize("protocol_text_kind", ["str", "bytes"])
@pytest.mark.parametrize("filename", ["real", "none"])
def test_parse_json_details(
    get_json_protocol_fixture: Callable[..., Any],
    fixture_version: str,
    fixture_name: str,
    protocol_text_kind: Literal["str", "bytes"],
    filename: str,
) -> None:
    protocol = get_json_protocol_fixture(
        fixture_version=fixture_version, fixture_name=fixture_name, decode=False
    )
    if protocol_text_kind == "str":
        protocol_text: Union[bytes, str] = protocol
    else:
        protocol_text = protocol.encode("utf-8")
    if filename == "real":
        fname: Optional[str] = "simple.json"
    else:
        fname = None
    parsed = parse(protocol_text, fname)
    assert isinstance(parsed, JsonProtocol)
    assert parsed.filename == fname
    assert parsed.contents == json.loads(protocol)
    assert parsed.metadata == parsed.contents["metadata"]
    assert parsed.schema_version == int(fixture_version)
    # TODO(IL, 2020-10-07): if schema v6 declares its own api_level,
    # then those v6 fixtures will need to be asserted differently
    assert parsed.api_level == API_VERSION_FOR_JSON_V5_AND_BELOW


def test_parse_bundle_details(get_bundle_fixture: Callable[..., Any]) -> None:
    fixture = get_bundle_fixture("simple_bundle")
    filename = fixture["filename"]

    parsed = parse(fixture["binary_zipfile"], filename)

    assert isinstance(parsed, PythonProtocol)
    assert parsed.filename == filename
    assert parsed.bundled_labware == fixture["bundled_labware"]
    assert parsed.bundled_python == fixture["bundled_python"]
    assert parsed.bundled_data == fixture["bundled_data"]
    assert parsed.metadata == fixture["metadata"]
    assert parsed.api_level == APIVersion(2, 0)


@pytest.mark.parametrize("protocol_file", ["testosaur_v2.py"])
def test_parse_extra_contents(
    get_labware_fixture: Callable[..., Any], protocol_file: str, protocol: Protocol
) -> None:
    fixture_96_plate = get_labware_fixture("fixture_96_plate")
    bundled_labware = {"fixture/fixture_96_plate/1": fixture_96_plate}
    extra_data = {"hi": b"there"}
    parsed = parse(
        protocol.text,
        "testosaur_v2.py",
        extra_labware=bundled_labware,
        extra_data=extra_data,
    )
    assert parsed.extra_labware == bundled_labware  # type: ignore[union-attr]
    assert parsed.bundled_data == extra_data  # type: ignore[union-attr]


@pytest.mark.parametrize(
    ("bad_protocol", "expected_message"),
    [
        # Bad Python syntax:
        (
            """
            metadata = {"apiLevel": "2.0"}
            def run(ctx)  # Missing ":".
                pass
            """,
            '(invalid syntax)|(Missing ":")',  # This error message depends on the Python version.
        ),
        # Bad Python Protocol API structure:
        (
            """
            metadata={"apiLevel": "2.0"}
            def run(ctx): pass
            def run(ctx): pass
            """,
            "More than one run function",
        ),
        (
            """
            metadata = {"apiLevel": "2.0"}

            print('hi')
            """,
            "No function 'run\\(ctx\\)'",  # This is a regex, so '\\' to escape '(' and ')' chars.
        ),
        (
            """
            metadata = {"apiLevel": "2.0"}
            def run(ctx):
              pass

            def run(blahblah):
              pass
            """,
            "More than one run function",
        ),
        # Various kinds of invalid metadata dict or apiLevel:
        (
            # metadata missing entirely.
            """
            def run():
                pass
            """,
            "apiLevel not declared",
        ),
        (
            # Metadata provided, but not as a dict.
            """
            metadata = "Hello"

            def run():
                pass
            """,
            "apiLevel not declared",
        ),
        (
            # apiLevel missing from metadata dict and requirements dict.
            """
            metadata = {"Hello": "World"}

            def run():
                pass
            """,
            "apiLevel not declared",
        ),
        (
            # Metadata not statically parsable.
            """
            metadata = {"apiLevel": "123" + ".456"}

            def run():
                pass
            """,
            "Could not read the contents of the metadata dict",
        ),
        (
            # apiLevel provided, but not as a string.
            """
            metadata = {"apiLevel": 123.456}

            def run():
                pass
            """,
            "must be strings",
        ),
        (
            # apiLevel provided, but not as a well formatted string.
            """
            metadata = {"apiLevel": "123*456"}

            def run():
                pass
            """,
            "is incorrectly formatted",
        ),
        (
            # robotType provided, but not a valid string.
            """
            metadata = {"apiLevel": "2.11"}
            requirements = {"robotType": "ot2"}

            def run():
                pass
            """,
            "robotType must be 'OT-2' or 'Flex', not 'ot2'.",
        ),
        (
            # robotType provided, but not a valid string.
            """
            metadata = {"apiLevel": "2.11"}
            requirements = {"robotType": "flex"}

            def run():
                pass
            """,
            "robotType must be 'OT-2' or 'Flex', not 'flex'.",
        ),
    ],
)
def test_parse_bad_structure(bad_protocol: str, expected_message: str) -> None:
    with pytest.raises(MalformedPythonProtocolError, match=expected_message):
        parse(dedent(bad_protocol))


# TODO(mm, 2023-08-10): When we remove python_parse_mode from parse(), remove this
# parametrization and merge these tests with the other metadata/requirements validation tests.
@pytest.mark.parametrize("python_parse_mode", PythonParseMode)
@pytest.mark.parametrize(
    ("questionable_protocol", "expected_message"),
    [
        (
            # apiLevel in both metadata and requirements.
            """
            metadata = {"apiLevel": "2.14"}
            requirements = {"apiLevel": "2.14"}
            def run(ctx): pass
            """,
            "You may only put apiLevel in the metadata dict or the requirements dict, not both.",
        ),
        (
            # apiLevel in both metadata and requirements.
            """
            metadata = {"apiLevel": "2.14"}
            requirements = {"apiLevel": ""}
            def run(ctx): pass
            """,
            "You may only put apiLevel in the metadata dict or the requirements dict, not both.",
        ),
        (
            # Unrecognized keys in requirements.
            """
            requirements = {
                "apiLevel": "2.15",
                "robotType": "Flex",
                "APILevel": "2.15",
                "RobotType": "Flex",
                "foo": "bar",
            }
            def run(ctx): pass
            """,
            "Unrecognized keys in requirements dict: 'APILevel', 'RobotType', 'foo'",
        ),
        (
            # apiLevel too old to support the Flex.
            """
            requirements = {"apiLevel": "2.13", "robotType": "Flex"}
            def run(ctx): pass
            """,
            "The Opentrons Flex only supports apiLevel 2.15 or newer.",
        ),
    ],
)
def test_errors_conditional_on_legacy_mode(
    questionable_protocol: str,
    python_parse_mode: PythonParseMode,
    expected_message: str,
) -> None:
    if python_parse_mode == PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS:
        # Should not raise:
        parse(dedent(questionable_protocol), python_parse_mode=python_parse_mode)
    else:
        with pytest.raises(MalformedPythonProtocolError, match=expected_message):
            parse(dedent(questionable_protocol), python_parse_mode=python_parse_mode)


# TODO(mm, 2023-08-10): Remove these tests when we remove python_parse_mode from parse().
# https://opentrons.atlassian.net/browse/RSS-306
@pytest.mark.parametrize(
    ("protocol_source", "expected_api_level"),
    [
        (
            """
            metadata = {"apiLevel": "2.15"}
            requirements = {"apiLevel": "2.14"}
            def run(ctx): pass
            """,
            APIVersion(2, 14),
        ),
        (
            """
            requirements = {"apiLevel": "2.14"}
            metadata = {"apiLevel": "2.15"}
            def run(ctx): pass
            """,
            APIVersion(2, 14),
        ),
    ],
)
def test_legacy_apilevel_override(
    protocol_source: str, expected_api_level: APIVersion
) -> None:
    """An apiLevel in requirements should override an apiLevel in metadata.

    This only matters with `PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS`.
    With stricter validation, it's impossible to put apiLevel in both dicts in the first place.
    """
    parsed = parse(
        dedent(protocol_source),
        python_parse_mode=PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS,
    )
    assert parsed.api_level == expected_api_level



import json
import os
from pathlib import Path
from typing import Callable

from opentrons_shared_data.labware.types import LabwareDefinition as LabwareDefDict

from opentrons.util.entrypoint_util import (
    FoundLabware,
    datafiles_from_paths,
    labware_from_paths,
)


def test_labware_from_paths(
    tmp_path: Path,
    get_labware_fixture: Callable[[str], LabwareDefDict],
) -> None:
    path_1 = tmp_path / "path 1"
    path_1.mkdir()
    path_2 = tmp_path / "path 2"
    path_2.mkdir()

    lw1 = get_labware_fixture("fixture_96_plate")
    lw2 = get_labware_fixture("fixture_24_tuberack")
    lw3 = get_labware_fixture("fixture_irregular_example_1")

    with open(path_1 / "labware1.json", "w") as lwtemp:
        json.dump(lw1, lwtemp)
    with open(path_1 / "labware2.json", "w") as lwtemp:
        json.dump(lw2, lwtemp)
    with open(path_2 / "labware3.json", "w") as lwtemp:
        json.dump(lw3, lwtemp)
    with open(path_2 / "invalid.json", "w") as lwtemp:
        lwtemp.write("asdjkashdkajvka")
    with open(path_2 / "notevenjson", "w") as lwtemp:
        lwtemp.write("bgbbabcba")
    with open(path_2 / "not_an_object.json", "w") as lwtemp:
        lwtemp.write('["hello", "world"]')

    res = labware_from_paths(
        [
            os.path.realpath(path_1),
            path_2,
        ]
    )
    assert res == {
        "fixture/fixture_96_plate/1": FoundLabware(path_1 / "labware1.json", lw1),
        "fixture/fixture_24_tuberack/1": FoundLabware(path_1 / "labware2.json", lw2),
        "fixture/fixture_irregular_example_1/1": FoundLabware(
            path_2 / "labware3.json", lw3
        ),
    }


def test_datafiles_from_paths(tmp_path: Path) -> None:
    os.mkdir(os.path.join(tmp_path, "path-1"))
    os.mkdir(os.path.join(tmp_path, "path-2"))
    with open(os.path.join(tmp_path, "path-1", "test3"), "wb") as f:
        f.write("oh hey there checkitout".encode("utf-8"))
    with open(os.path.join(tmp_path, "path-2", "test2"), "wb") as f:
        f.write("oh man this isnt even utf8".encode("utf-16"))
    with open(os.path.join(tmp_path, "path-2", "test1"), "wb") as f:
        f.write("wait theres a second file???".encode())
    with open(os.path.join(tmp_path, "test-file"), "wb") as f:
        f.write("this isnt even in a directory".encode())

    res = datafiles_from_paths(
        [
            os.path.join(tmp_path, "path-1"),
            Path(os.path.join(tmp_path, "path-2")),
            os.path.join(tmp_path, "test-file"),
        ]
    )
    assert res == {
        "test3": "oh hey there checkitout".encode("utf-8"),
        "test2": "oh man this isnt even utf8".encode("utf-16"),
        "test1": "wait theres a second file???".encode(),
        "test-file": "this isnt even in a directory".encode(),
    }



"""Test evotip dispense in place commands."""

import pytest
from decoy import Decoy

from opentrons_shared_data.labware import load_definition
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
    labware_definition_type_adapter,
)

from opentrons.protocol_engine import (
    DeckPoint,
    LiquidHandlingWellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.pressure_dispense import (
    PressureDispenseImplementation,
    PressureDispenseParams,
    PressureDispenseResult,
)
from opentrons.protocol_engine.execution import (
    GantryMover,
    MovementHandler,
    PipettingHandler,
)
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LabwareWellId
from opentrons.types import Point


@pytest.fixture
def evotips_definition() -> LabwareDefinition:
    """A fixturee of the evotips definition."""
    # TODO (chb 2025-01-29): When we migrate all labware to v3 we can clean this up
    return labware_definition_type_adapter.validate_python(
        load_definition("ev_resin_tips_flex_96_labware", 1)
    )


@pytest.fixture
def subject(
    pipetting: PipettingHandler,
    state_view: StateView,
    gantry_mover: GantryMover,
    model_utils: ModelUtils,
    movement: MovementHandler,
    **kwargs: object,
) -> PressureDispenseImplementation:
    """Build a command implementation."""
    return PressureDispenseImplementation(
        pipetting=pipetting,
        state_view=state_view,
        gantry_mover=gantry_mover,
        model_utils=model_utils,
        movement=movement,
    )


async def test_pressure_dispense_implementation(
    decoy: Decoy,
    movement: MovementHandler,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    state_view: StateView,
    subject: PressureDispenseImplementation,
    evotips_definition: LabwareDefinition,
) -> None:
    """It should dispense in place."""
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.TOP, offset=WellOffset(x=0, y=0, z=0)
    )

    data = PressureDispenseParams(
        pipetteId="pipette-id-abc123",
        labwareId="labware-id-abc123",
        wellName="A3",
        volume=100,
        flowRate=456,
    )

    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id-abc123",
            labware_id="labware-id-abc123",
            well_name="A3",
            well_location=well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(state_view.labware.get_definition("labware-id-abc123")).then_return(
        evotips_definition
    )
    decoy.when(
        state_view.pipettes.get_aspirated_volume("pipette-id-abc123")
    ).then_return(100)

    decoy.when(pipetting.get_state_view()).then_return(state_view)

    decoy.when(
        await pipetting.dispense_in_place(
            pipette_id="pipette-id-abc123",
            volume=100.0,
            flow_rate=456.0,
            push_out=None,
            is_full_dispense=True,
            correction_volume=0,
        )
    ).then_return(100)

    decoy.when(await gantry_mover.get_position("pipette-id-abc123")).then_return(
        Point(1, 2, 3)
    )

    result = await subject.execute(data)

    assert result == SuccessData(
        public=PressureDispenseResult(volume=100),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id-abc123",
                new_location=LabwareWellId(
                    labware_id="labware-id-abc123",
                    well_name="A3",
                ),
                new_deck_point=DeckPoint.model_construct(x=1, y=2, z=3),
            ),
            pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
                pipette_id="pipette-id-abc123", volume=100
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="pipette-id-abc123", ready_to_aspirate=False
            ),
        ),
    )
    decoy.verify(await pipetting.increase_evo_disp_count("pipette-id-abc123"), times=1)



"""Tests for opentrons.legacy_commands.publisher."""

from __future__ import annotations

from typing import Any, Dict, cast

import pytest
from decoy import Decoy, matchers

from opentrons.legacy_broker import LegacyBroker
from opentrons.legacy_commands.publisher import (
    CommandPublisher,
    publish,
    publish_context,
)
from opentrons.legacy_commands.types import Command as CommandDict
from opentrons.legacy_commands.types import CommandMessage


@pytest.fixture
def broker(decoy: Decoy) -> LegacyBroker:
    """Return a mocked out Broker."""
    return decoy.mock(cls=LegacyBroker)


def test_publish_decorator(decoy: Decoy, broker: LegacyBroker) -> None:
    """It should publish "before" and "after" messages for decorated methods."""
    _act = decoy.mock(name="_act")

    def _get_command_payload(foo: str, bar: int) -> CommandDict:
        return cast(
            CommandDict,
            {"name": "some_command", "payload": {"foo": foo, "bar": bar}},
        )

    class _Subject(CommandPublisher):
        @publish(command=_get_command_payload)
        def act(self, foo: str, bar: int) -> None:
            _act()

    subject = _Subject(broker=broker)
    subject.act("hello", 42)

    before_message_id = matchers.Captor()
    after_message_id = matchers.Captor()

    decoy.verify(
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "before",
                    "id": before_message_id,
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
        _act(),
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "after",
                    "id": after_message_id,
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
    )

    assert before_message_id.value == after_message_id.value


def test_publish_decorator_with_arg_defaults(
    decoy: Decoy, broker: LegacyBroker
) -> None:
    """It should pass method argument defaults to the command creator."""
    _act = decoy.mock(name="_act")

    def _get_command_payload(foo: str, bar: int) -> CommandDict:
        return cast(
            CommandDict,
            {"name": "some_command", "payload": {"foo": foo, "bar": bar}},
        )

    class _Subject(CommandPublisher):
        @publish(command=_get_command_payload)
        def act(self, foo: str, bar: int = 42) -> None:
            _act()

    subject = _Subject(broker=broker)
    subject.act("hello")

    decoy.verify(
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "before",
                    "id": matchers.IsA(str),
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
        _act(),
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "after",
                    "id": matchers.IsA(str),
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
    )


def test_publish_decorator_with_error(decoy: Decoy, broker: LegacyBroker) -> None:
    """It should capture an exception and place it in the "after" message."""

    def _get_command_payload(foo: str, bar: int) -> Dict[str, Any]:
        return {"name": "some_command", "payload": {"foo": foo, "bar": bar}}

    class _Subject(CommandPublisher):
        @publish(command=_get_command_payload)  # type: ignore[arg-type]
        def act(self, foo: str, bar: int) -> None:
            raise RuntimeError("oh no")

    subject = _Subject(broker=broker)

    with pytest.raises(RuntimeError, match="oh no"):
        subject.act("hello", 42)

    before_message_id = matchers.Captor()
    after_message_id = matchers.Captor()

    decoy.verify(
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "before",
                    "id": before_message_id,
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "after",
                    "id": after_message_id,
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": matchers.IsA(RuntimeError),
                },
            ),
        ),
    )

    assert before_message_id.value == after_message_id.value


def test_publish_decorator_remaps_instrument(
    decoy: Decoy, broker: LegacyBroker
) -> None:
    """It should pass "self" to command creator arguments named "instrument"."""
    _act = decoy.mock(name="_act")

    def _get_command_payload(foo: str, instrument: _Subject) -> Dict[str, Any]:
        return {
            "name": "some_command",
            "payload": {"foo": foo, "bar": instrument.bar},
        }

    class _Subject(CommandPublisher):
        bar: int = 42

        @publish(command=_get_command_payload)  # type: ignore[arg-type]
        def act(self, foo: str) -> None:
            _act()

    subject = _Subject(broker=broker)

    subject.act("hello")

    decoy.verify(
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "before",
                    "id": matchers.IsA(str),
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
        _act(),
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "after",
                    "id": matchers.IsA(str),
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
    )


def test_publish_context(decoy: Decoy, broker: LegacyBroker) -> None:
    _act = decoy.mock(name="_act")

    command = cast(
        CommandDict,
        {"name": "some_command", "payload": {"foo": "hello", "bar": 42}},
    )

    def _published_func(foo: str, bar: int) -> None:
        _act()

    with publish_context(broker=broker, command=command):
        _published_func("hello", 42)

    before_message_id = matchers.Captor()
    after_message_id = matchers.Captor()

    decoy.verify(
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "before",
                    "id": before_message_id,
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
        _act(),
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "after",
                    "id": after_message_id,
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
    )

    assert before_message_id.value == after_message_id.value


def test_publish_context_with_error(decoy: Decoy, broker: LegacyBroker) -> None:
    command = cast(
        CommandDict,
        {"name": "some_command", "payload": {"foo": "hello", "bar": 42}},
    )

    def _published_func(foo: str, bar: int) -> None:
        raise RuntimeError("oh no")

    with pytest.raises(RuntimeError, match="oh no"):
        with publish_context(broker=broker, command=command):
            _published_func("hello", 42)

    before_message_id = matchers.Captor()
    after_message_id = matchers.Captor()

    decoy.verify(
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "before",
                    "id": before_message_id,
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": None,
                },
            ),
        ),
        broker.publish(
            topic="command",
            message=cast(
                CommandMessage,
                {
                    "$": "after",
                    "id": after_message_id,
                    "name": "some_command",
                    "payload": {"foo": "hello", "bar": 42},
                    "error": matchers.IsA(RuntimeError),
                },
            ),
        ),
    )

    assert before_message_id.value == after_message_id.value



import json

import pytest

from opentrons.calibration_storage import helpers, types


def test_hash_labware_def() -> None:
    def1a = {"metadata": {"a": 123}, "importantStuff": [1.1, 0.00003, 1 / 3]}
    def1aa = {"metadata": {"a": 123}, "importantStuff": [1.1, 0.00003, 1 / 3]}
    def1b = {"metadata": {"a": "blah"}, "importantStuff": [1.1, 0.00003, 1 / 3]}
    def2 = {"metadata": {"a": 123}, "importantStuff": [1.1, 0.000033, 1 / 3]}

    # identity preserved across json serialization+deserialization
    assert helpers.hash_labware_def(def1a) == helpers.hash_labware_def(  # type: ignore[arg-type]
        json.loads(json.dumps(def1a, separators=(",", ":")))
    )
    # 2 instances of same def should match
    assert helpers.hash_labware_def(def1a) == helpers.hash_labware_def(def1aa)  # type: ignore[arg-type]
    # metadata ignored
    assert helpers.hash_labware_def(def1a) == helpers.hash_labware_def(def1b)  # type: ignore[arg-type]
    # different data should not match
    assert helpers.hash_labware_def(def1a) != helpers.hash_labware_def(def2)  # type: ignore[arg-type]


@pytest.mark.parametrize(
    argnames=["uri", "expected"],
    argvalues=[
        [
            "myspace/labware_is_cool/1",
            types.UriDetails("myspace", "labware_is_cool", 1),
        ],
        ["", types.UriDetails("", "", 1)],
        ["my_totally_cool_labware", types.UriDetails("", "my_totally_cool_labware", 1)],
    ],
)
def test_details_from_uri(uri: str, expected: types.UriDetails) -> None:
    assert helpers.details_from_uri(uri) == expected
